from __future__ import annotations

import itertools
import math
from typing import TYPE_CHECKING, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndii
from mpl_toolkits.axes_grid1 import make_axes_locatable
from numpy.typing import DTypeLike, NDArray
from scipy.ndimage import affine_transform as affine_transform_scipy
from scipy.ndimage import fourier_shift as fourier_shift_scipy
from scipy.signal.windows import tukey as tukey_scipy
from skimage.registration import (
    phase_cross_correlation as phase_cross_correlation_skimage,
)

from spfluo.utils.array import (
    array_namespace,
    get_device,
    is_array_api_obj,
    is_cupy_array,
    is_numpy_array,
    is_torch_array,
    to_numpy,
)

if TYPE_CHECKING:
    from spfluo.utils.array import Array, array_api_module


def affine_transform(
    input: "Array",
    matrix: "Array",
    offset: Union[float, Tuple[float], "Array"] = 0.0,
    output_shape: Optional[Tuple[int]] = None,
    output: Optional[Union["Array", DTypeLike]] = None,
    order: int = 3,
    mode: str = "constant",
    cval: float = 0.0,
    prefilter: bool = True,
    *,
    batch: bool = False,
    multichannel: bool = False,
) -> "Array":
    """Apply affine transformations to an image.
    Works with multichannel images and batches.
    Supports numpy, cupy and torch inputs.
    torch only supports linear interpolation.

    Given an output image pixel index vector ``o``, the pixel value is
    determined from the input image at position
    ``xp.dot(matrix, o) + offset``.

    Args:
        input (xp.ndarray): The input array.
            torch only supports 3D inputs.
        matrix (xp.ndarray): Must be a 'real floating' dtype matrix.
            The inverse coordinate transformation matrix,
            mapping output coordinates to input coordinates. If ``ndim`` is the
            number of dimensions of ``input``, the given matrix must have one
            of the following shapes:

                - ``(N, ndim, ndim)``: the linear transformation matrix for each
                  output coordinate.
                - ``(N, ndim,)``: assume that the 2D transformation matrix is
                  diagonal, with the diagonal specified by the given value.
                - ``(N, ndim + 1, ndim + 1)``: assume that the transformation is
                  specified using homogeneous coordinates. In this case, any
                  value passed to ``offset`` is ignored.
                - ``(N, ndim, ndim + 1)``: as above, but the bottom row of a
                  homogeneous transformation matrix is always
                  ``[0, 0, ..., 1]``, and may be omitted.

        offset (float or sequence or cp.array): The offset into the array where
            the transform is applied. If a float, ``offset`` is the same for each
            axis. If a sequence, ``offset`` should contain one value for each
            axis. If a xp.array, should be of shape (N, d) where d is the number
            of axes.
        output_shape (tuple of ints): Shape tuple. One shape for all the batch.
        output (xp.ndarray or ~xp.dtype): The array in which to place the
            output, or the dtype of the returned array.
            Not implemented in torch.
        order (int): The order of the spline interpolation, default is 3. Must
            be in the range 0-5.
            Only order 1 is implemented in torch.
        mode (str): Points outside the boundaries of the input are filled
            according to the given mode (``'constant'``, ``'nearest'``,
            ``'mirror'``, ``'reflect'``, ``'wrap'``, ``'grid-mirror'``,
            ``'grid-wrap'``, ``'grid-constant'`` or ``'opencv'``).
            Only ``'constant'``, ``'nearest'``, ``'reflect'`` are implemented
            in torch.
        cval (scalar): Value used for points outside the boundaries of
            the input if ``mode='constant'`` or ``mode='opencv'``. Default is
            0.0.
            Only 0.0 is implemented in torch.
        prefilter (bool): Determines if the input array is prefiltered with
            ``spline_filter`` before interpolation. The default is True, which
            will create a temporary ``float64`` array of filtered values if
            ``order > 1``. If setting this to False, the output will be
            slightly blurred if ``order > 1``, unless the input is prefiltered,
            i.e. it is the result of calling ``spline_filter`` on the original
            input.
            Not implemented in torch.

        batch (bool): if True, the first dimension is a batch dimension
            default to False
        multichannel (bool): if True, the first (or second if batch=True) is
            the channel dimension

    Returns:
        xp.ndarray:
            The transformed input. Return None if output is given.
    """
    xp = array_namespace(input, matrix)
    assert xp.isdtype(
        matrix.dtype, "real floating"
    ), "matrix dtype must be a 'real floating' dtype (float32, float64)"
    has_output = False
    if is_array_api_obj(output):
        output = xp.asarray(output)
        has_output = True

    if batch is False:
        input = input[None, ...]
        matrix = matrix[None, ...]
        if has_output:
            output = output[None]
    if multichannel is False:
        input = input[:, None, ...]

    if is_torch_array(input):
        from ._torch_functions.volume import (
            affine_transform_batched_multichannel_pytorch,
        )

        func = affine_transform_batched_multichannel_pytorch
    elif is_cupy_array(input):
        from ._cupy_functions.volume import affine_transform_batched_multichannel_cupy

        func = affine_transform_batched_multichannel_cupy
    elif is_numpy_array(input):
        func = affine_transform_batched_multichannel_scipy
    else:
        try:
            input = np.asarray(input)
            matrix = np.asarray(matrix)
            offset = np.asarray(offset)
            if has_output:
                output = np.asarray(output)
        except TypeError:
            raise ValueError(f"No backend found for {xp}")
        func = affine_transform_batched_multichannel_scipy
    out = func(
        input, matrix, offset, output_shape, output, order, mode, cval, prefilter
    )
    if has_output:
        out = output
    out = xp.asarray(out)
    if multichannel is False:
        out = out[:, 0, ...]
    if batch is False:
        out = out[0, ...]
    if not has_output:
        return out


def affine_transform_batched_multichannel_scipy(
    input: NDArray,
    matrix: NDArray,
    offset: Union[float, Tuple[float], NDArray] = 0.0,
    output_shape: Optional[Tuple[int]] = None,
    output: Optional[Union[NDArray, DTypeLike]] = None,
    order: int = 1,
    mode: str = "constant",
    cval: float = 0.0,
    prefilter: bool = True,
) -> NDArray:
    N, C, *image_shape = input.shape
    if output_shape is None:
        output_shape = tuple(image_shape)
    return_none = False
    if output is None:
        output = np.empty((N, C) + output_shape, dtype=input.dtype)
    elif type(output) is type:
        output = np.empty((N, C) + output_shape, dtype=output)
    else:
        return_none = True
    if isinstance(offset, float) or isinstance(offset, tuple):

        def offset_gen(_):
            return offset

    else:
        assert type(offset) is np.ndarray
        if offset.ndim <= 1:
            assert offset.ndim == 0 or (offset.shape == (1,) or offset.shape == (3,))

            def offset_gen(_):
                return offset
        elif offset.ndim == 2:
            assert offset.shape[0] == N
            assert offset.shape[-1] == (1,) or offset.shape[-1] == (3,)

            def offset_gen(i):
                return offset[i]
        else:
            raise ValueError("Wrong offset size")

    for i in range(N):
        for j in range(C):
            affine_transform_scipy(
                input[i, j],
                matrix[i],
                offset_gen(i),
                output_shape,
                output[i, j],
                order,
                mode,
                cval,
                prefilter,
            )

    if return_none:
        return

    return output


def resample(
    volume: "Array",
    sampling: tuple[float],
    order: int = 1,
    batch: bool = False,
    multichannel: bool = False,
):
    xp = array_namespace(volume)
    device = get_device(volume)
    sampling = xp.asarray(sampling, device=device, dtype=xp.float64)
    in_shape = xp.asarray(volume.shape[-3:], device=device, dtype=xp.float64)
    out_shape = xp.round(in_shape * sampling)
    D, H, W = int(out_shape[0]), int(out_shape[1]), int(out_shape[2])

    # Compute transformation matrix
    input_center, output_center = (in_shape - 1) / 2, (out_shape - 1) / 2
    H_center, H_homo = (
        xp.eye(4, device=device, dtype=xp.float64),
        xp.eye(4, device=device, dtype=xp.float64),
    )
    H_center[:3, 3] = -input_center  # 1. translation to (0,0,0)
    H_homo[[0, 1, 2], [0, 1, 2]] = sampling  # 2. homothety
    H_homo[:3, 3] = output_center  # 3. translation to center of image
    #    3-2 <- 1
    mat = H_homo @ H_center

    inv_mat = xp.linalg.inv(mat)
    if batch:
        N = volume.shape[0]
        inv_mat = xp.broadcast_to(inv_mat[None], (N, 4, 4))
    return affine_transform(
        volume,
        inv_mat,
        output_shape=(D, H, W),
        batch=batch,
        multichannel=multichannel,
        order=order,
        prefilter=False,
    )


def pad(volume: "Array", pad_width: Tuple[int]):
    xp = array_namespace(volume)
    pad_width = np.asarray(pad_width)
    if pad_width.ndim == 1:
        pad_width = np.broadcast_to(np.reshape(pad_width, (-1, 1)), (volume.ndim, 2))
    assert pad_width.shape == (volume.ndim, 2)
    padded_volume = xp.zeros(
        tuple(
            [
                s + int(before) + int(after)
                for s, (before, after) in zip(volume.shape, pad_width)
            ]
        ),
        dtype=volume.dtype,
        device=get_device(volume),
    )
    padded_volume[
        tuple(
            [
                slice(before, s - after)
                for (before, after), s in zip(pad_width, padded_volume.shape)
            ]
        )
    ] = volume

    return padded_volume


def interpolate_to_size(
    volume: "Array",
    output_size: Tuple[int, int, int],
    order=1,
    batch=False,
    multichannel=False,
) -> "Array":
    """
    Used for padding. The zoom matrix will zoom-out from the image.
    """
    xp = array_namespace(volume)
    volume = xp.asarray(volume)
    D, H, W = output_size
    input_center = (
        xp.asarray(volume.shape[-3:], device=get_device(volume), dtype=xp.float64) - 1
    ) / 2
    output_center = (
        xp.asarray(output_size, device=get_device(volume), dtype=xp.float64) - 1
    ) / 2
    mat = xp.eye(4, device=get_device(volume))
    mat[:3, 3] = output_center - input_center
    inv_mat = xp.linalg.inv(mat)
    if batch:
        N = volume.shape[0]
        inv_mat = xp.broadcast_to(inv_mat[None, ...], (N, 4, 4))
    out_vol = affine_transform(
        volume,
        inv_mat,
        output_shape=(D, H, W),
        batch=batch,
        multichannel=multichannel,
        order=order,
        prefilter=False,
    )
    return out_vol


def fourier_shift_broadcasted_scipy(
    input: NDArray,
    shift: Union[float, Sequence[float], NDArray],
    n: int = -1,
    axis: int = -1,
    output: Optional[NDArray] = None,
):
    shift = np.asarray(shift)
    if shift.ndim == 0:
        shift = np.asarray([shift] * input.ndim)
    nb_spatial_dims = shift.shape[-1]
    broadcasted_shape = np.broadcast_shapes(
        input.shape[:-nb_spatial_dims], shift.shape[:-1]
    )
    image_shape = input.shape[-nb_spatial_dims:]
    input = np.broadcast_to(input, broadcasted_shape + image_shape)
    shift = np.broadcast_to(shift, broadcasted_shape + (nb_spatial_dims,))
    output = np.empty(broadcasted_shape + image_shape, dtype=input.dtype)
    for index in np.ndindex(broadcasted_shape):
        fourier_shift_scipy(
            input[index].copy(),
            shift[index],
            n,
            axis,
            output[index],
        )
    return output


def fourier_shift(
    input: "Array",
    shift: Union[float, Sequence[float], "Array"],
    n: int = -1,
    axis: int = -1,
    output: Optional["Array"] = None,
):
    """
    Multidimensional Fourier shift filter.

    The array is multiplied with the Fourier transform of a shift operation.

    Parameters
    ----------
    input : array_like
        The input array.
        If shift is an array, input and shift will be broadcasted:
            input of shape ({...}, [...])
            where [...] corresponds to the D spatial dimensions
            and {...} corresponds to the dimensions to be broadcasted
    shift : float, sequence or array_like
        The size of the box used for filtering.
        If a float, `shift` is the same for all axes. If a sequence, `shift`
        has to contain one value for each axis.
        If an array, shift will be broadcasted with the input :
            shift must be of shape ({{...}}, D)
            where {{...}} corresponds to dimensions to be broadcasted
            and D to the number of spatial dimensions
    n : int, optional
        If `n` is negative (default), then the input is assumed to be the
        result of a complex fft.
        If `n` is larger than or equal to zero, the input is assumed to be the
        result of a real fft, and `n` gives the length of the array before
        transformation along the real transform direction.
    axis : int, optional
        The axis of the real transform.
    output : ndarray, optional
        If given, the result of shifting the input is placed in this array.
        None is returned in this case.
    Returns
    -------
    fourier_shift : ndarray
        The shifted input.
        If shift is an array, {...} and {{...}} are broadcasted to (...).
        The resulting shifted array has the shape ((...), [...])
    """
    if is_numpy_array(input):
        func = fourier_shift_broadcasted_scipy
    elif is_cupy_array(input):
        from ._cupy_functions.volume import fourier_shift_broadcasted_cupy

        func = fourier_shift_broadcasted_cupy
    else:
        func = _fourier_shift_broadcasted_array_api

    output = func(
        input,
        shift,
        n,
        axis,
        output,
    )
    return output


def _fourier_shift_broadcasted_array_api(
    input: "Array",
    shift: Union[float, Sequence[float], "Array"],
    n: int = -1,
    axis: int = -1,
    output: "Optional[Array]" = None,
):
    """
    Args:
        input (Array): input in the Fourier domain ({...}, [...])
            where [...] corresponds to the N spatial dimensions
            and {...} corresponds to the batched dimensions
        shift (Array): shift to apply to the input ({{...}}, N)
            where {{...}} corresponds to batched dimensions.
        n: not implemented
        axis: not implemented
        output: not implemented
    Notes:
        {...} and {{...}} are broadcasted to (...).
    Returns:
        out (Array): input shifted in the Fourier domain. Shape ((...), [...])
    """
    xp = array_namespace(input)
    if n != -1:
        raise NotImplementedError("n should be equal to -1")
    if axis != -1:
        raise NotImplementedError("axis should be equal to -1")
    if output is not None:
        raise NotImplementedError("can't store result in output. not implemented")
    device = get_device(input)
    complex_dtype = input.dtype
    assert complex_dtype in (xp.complex64, xp.complex128)
    floating_dtype = None
    if input.dtype == xp.complex128:
        floating_dtype = xp.float64
    else:
        floating_dtype = xp.float32
    shift = xp.asarray(shift, dtype=floating_dtype, device=device)
    if shift.ndim == 0:
        shift = xp.asarray([shift] * input.ndim)
    nb_spatial_dims = shift.shape[-1]
    spatial_shape = input.shape[-nb_spatial_dims:]
    shift = xp.reshape(shift, (*shift.shape[:-1], *((1,) * nb_spatial_dims), -1))

    grid_freq = xp.stack(
        xp.meshgrid(
            *[xp.asarray(np.fft.fftfreq(int(s)), device=device) for s in spatial_shape],
            indexing="ij",
        ),
        axis=-1,
    )
    phase_shift = xp.sum(grid_freq * shift, axis=-1)

    # Fourier shift
    out = input * xp.exp(
        -1j * 2 * xp.pi * xp.astype(phase_shift, complex_dtype, copy=False)
    )

    return out


def phase_cross_correlation_broadcasted_skimage(
    reference_image: "Array",
    moving_image: "Array",
    *,
    upsample_factor: int = 1,
    space: str = "real",
    disambiguate: bool = False,
    reference_mask: Optional["Array"] = None,
    moving_mask: Optional["Array"] = None,
    overlap_ratio: float = 0.3,
    normalization: str = "phase",
    nb_spatial_dims: Optional[int] = None,
):
    if nb_spatial_dims is None:
        return phase_cross_correlation_skimage(
            reference_image,
            moving_image,
            upsample_factor=upsample_factor,
            space=space,
            disambiguate=disambiguate,
            reference_mask=reference_mask,
            moving_mask=moving_mask,
            overlap_ratio=overlap_ratio,
            normalization=normalization,
        )
    broadcast = np.broadcast(reference_image, moving_image)
    reference_image, moving_image = np.broadcast_arrays(reference_image, moving_image)
    other_shape = broadcast.shape[:-nb_spatial_dims]
    shifts = [np.empty(other_shape) for _ in range(nb_spatial_dims)]
    errors = np.empty(other_shape)
    phasediffs = np.empty(other_shape)
    for index in np.ndindex(other_shape):
        ref_im, moving_im = reference_image[index], moving_image[index]
        s, e, p = phase_cross_correlation_skimage(
            ref_im,
            moving_im,
            upsample_factor=upsample_factor,
            space=space,
            disambiguate=disambiguate,
            reference_mask=reference_mask,
            moving_mask=moving_mask,
            overlap_ratio=overlap_ratio,
            normalization=normalization,
        )
        for i in range(nb_spatial_dims):
            shifts[i][index] = s[i]
        errors[index] = e
        phasediffs[index] = p
    return tuple(shifts), errors, phasediffs


def phase_cross_correlation(
    reference_image: "Array",
    moving_image: "Array",
    *,
    upsample_factor: int = 1,
    space: str = "real",
    disambiguate: bool = False,
    reference_mask: Optional["Array"] = None,
    moving_mask: Optional["Array"] = None,
    overlap_ratio: float = 0.3,
    normalization: str = "phase",
    nb_spatial_dims: Optional[int] = None,
    multichannel: bool = False,
):
    """Efficient subpixel image translation registration by cross-correlation.

    This code gives the same precision as the FFT upsampled cross-correlation
    in a fraction of the computation time and with reduced memory requirements.
    It obtains an initial estimate of the cross-correlation peak by an FFT and
    then refines the shift estimation by upsampling the DFT only in a small
    neighborhood of that estimate by means of a matrix-multiply DFT [1]_.

    Parameters
    ----------
    reference_image : array
        Reference image.
    moving_image : array
        Image to register. Must be same dimensionality as
        ``reference_image``.
    upsample_factor : int, optional
        Upsampling factor. Images will be registered to within
        ``1 / upsample_factor`` of a pixel. For example
        ``upsample_factor == 20`` means the images will be registered
        within 1/20th of a pixel. Default is 1 (no upsampling).
        Not used if any of ``reference_mask`` or ``moving_mask`` is not None.
    space : string, one of "real" or "fourier", optional
        Defines how the algorithm interprets input data. "real" means
        data will be FFT'd to compute the correlation, while "fourier"
        data will bypass FFT of input data. Case insensitive. Not
        used if any of ``reference_mask`` or ``moving_mask`` is not
        None.
    disambiguate : bool
        The shift returned by this function is only accurate *modulo* the
        image shape, due to the periodic nature of the Fourier transform. If
        this parameter is set to ``True``, the *real* space cross-correlation
        is computed for each possible shift, and the shift with the highest
        cross-correlation within the overlapping area is returned.
    reference_mask : ndarray
        Boolean mask for ``reference_image``. The mask should evaluate
        to ``True`` (or 1) on valid pixels. ``reference_mask`` should
        have the same shape as ``reference_image``.
    moving_mask : ndarray or None, optional
        Boolean mask for ``moving_image``. The mask should evaluate to ``True``
        (or 1) on valid pixels. ``moving_mask`` should have the same shape
        as ``moving_image``. If ``None``, ``reference_mask`` will be used.
    overlap_ratio : float, optional
        Minimum allowed overlap ratio between images. The correlation for
        translations corresponding with an overlap ratio lower than this
        threshold will be ignored. A lower `overlap_ratio` leads to smaller
        maximum translation, while a higher `overlap_ratio` leads to greater
        robustness against spurious matches due to small overlap between
        masked images. Used only if one of ``reference_mask`` or
        ``moving_mask`` is not None.
    normalization : {"phase", None}
        The type of normalization to apply to the cross-correlation. This
        parameter is unused when masks (`reference_mask` and `moving_mask`) are
        supplied.
    nb_spatial_dims: int
        If your inputs are broadcastable, you must fill this param.
    multichannel: bool
        if True, reference_image and moving_image must have shape ({...}, C, [...])
        where [...] corresponds to the spatial dims

    Returns
    -------
    shift : array
        Shift vector (in pixels) required to register ``moving_image``
        with ``reference_image``. Axis ordering is consistent with
        the axis order of the input array.
    error : float
        Translation invariant normalized RMS error between
        ``reference_image`` and ``moving_image``.
        is "always".
    phasediff : float
        Global phase difference between the two images (should be
        zero if images are non-negative).
    """
    extra_kwargs = {}
    if multichannel:
        func = _phase_cross_correlation_broadcasted_array_api
        extra_kwargs["multichannel"] = True
    elif is_numpy_array(reference_image):
        func = phase_cross_correlation_broadcasted_skimage
    elif is_cupy_array(reference_image):
        from ._cupy_functions.volume import phase_cross_correlation_broadcasted_cucim

        func = phase_cross_correlation_broadcasted_cucim
    else:
        func = _phase_cross_correlation_broadcasted_array_api

    shift, error, phasediff = func(
        reference_image,
        moving_image,
        upsample_factor=upsample_factor,
        space=space,
        disambiguate=disambiguate,
        reference_mask=reference_mask,
        moving_mask=moving_mask,
        overlap_ratio=overlap_ratio,
        normalization=normalization,
        nb_spatial_dims=nb_spatial_dims,
        **extra_kwargs,
    )

    return shift, error, phasediff


def swap_dims(arr: "Array", axis1: int, axis2: int):
    xp = array_namespace(arr)
    if axis1 < 0:
        axis1 += arr.ndim
    if axis2 < 0:
        axis2 += arr.ndim
    axes = list(range(arr.ndim))
    i1, i2 = axes.index(axis1), axes.index(axis2)
    axes[i1], axes[i2] = axes[i2], axes[i1]
    return xp.permute_dims(arr, tuple(axes))


def _upsampled_dft(
    data: "Array",
    upsampled_region_size: int,
    upsample_factor: int = 1,
    axis_offsets: "Array|None" = None,
    nb_spatial_dims=3,
):
    xp = array_namespace(data)
    device = get_device(data)
    if axis_offsets is None:
        raise NotImplementedError()
    upsampled_region_size = [upsampled_region_size] * nb_spatial_dims
    dim_properties = list(
        zip(
            range(-nb_spatial_dims, 0),
            data.shape[-nb_spatial_dims:],
            upsampled_region_size,
            xp.permute_dims(axis_offsets, (-1, *tuple(range(axis_offsets.ndim - 1)))),
        )
    )
    im2pi = 1j * 2 * xp.pi
    for ax, n_items, ups_size, ax_offset in dim_properties[::-1]:
        kernel = (
            xp.arange(ups_size, device=device, dtype=xp.float32) - ax_offset[..., None]
        )[..., None] * xp.asarray(
            np.fft.fftfreq(n_items, d=upsample_factor), device=device
        )
        kernel = xp.exp(-im2pi * xp.astype(kernel, data.dtype))  # shape (..., U, s_i)
        for _ in range(nb_spatial_dims - 1):
            kernel = xp.expand_dims(kernel, axis=-3)
        kernel  # shape (..., 1, .., 1, U, s_i)
        data  # shape (..., s_1, .., s_i .., s_N)
        data = swap_dims(data, ax, -1)  # shape (..., s_1, .., s_N, .., s_i)
        data = data[..., None]  # shape (..., s_1, .., s_N, .., s_i, 1)
        data = (kernel @ data)[..., 0]  # shape (..., s_1, .., s_N, .., U)
        data = swap_dims(data, ax, -1)
    return data


def unravel_index(indices: "Array", shape: tuple[int, ...]):
    xp = array_namespace(indices)
    device = get_device(indices)
    dim_prod = np.flip(np.cumprod((1, *shape[:0:-1])))
    dim_prod = xp.asarray(dim_prod.tolist(), dtype=indices.dtype, device=device)
    return (indices[..., None] // dim_prod) % xp.asarray(
        shape, dtype=indices.dtype, device=device
    )


def _cross_correlation_multichannel_max(
    x: "Array", y: "Array", normalization: str, nb_spatial_dims: int = None
) -> "tuple[Array, Array, Array]":
    """Compute multichannel cross-correlation between x and y
    Params:
        x (Array) of shape ({...}, C, ...)
            where (...) corresponds to the N spatial dimensions
            and C is the number of channels
        y (Array) of shape ({{...}}, C, ...)
            where (...) corresponds to the N spatial dimensions
    Returns:
        maxi (Array): cross correlation maximum of shape [...]
        shift (Tuple[Array]): tuple of N tensors of size [...]
        image_product (Array): product of size ([...], ...)
    Notes
        x and y must be broadcastable
        {...} and {{...}} are broadcasted to [...]
    """
    xp = array_namespace(x, y)
    nb_spatial_dims = nb_spatial_dims if nb_spatial_dims is not None else x.ndim - 1
    output_shape = np.broadcast_shapes(x.shape, y.shape)
    other_shape, spatial_shape = (
        output_shape[: -nb_spatial_dims - 1],
        output_shape[-nb_spatial_dims:],
    )
    z = x * xp.conj(y)
    # reduce channel dim
    z = xp.sum(z, axis=-nb_spatial_dims - 1)
    if normalization == "phase":
        z_abs = xp.abs(z)
        eps = xp.finfo(z_abs.dtype).eps
        z /= xp.where(
            z_abs > 100 * eps,
            xp.asarray(z_abs, dtype=z.dtype),
            xp.asarray(100 * eps, dtype=z.dtype),
        )
    cc = xp.fft.ifftn(z, axes=tuple(range(-nb_spatial_dims, 0)))
    cc = xp.abs(cc)
    cc = xp.reshape(cc, other_shape + (-1,))
    max_idx = xp.argmax(cc, axis=-1)
    indices = tuple(
        xp.asarray(idx, dtype=max_idx.dtype, device=get_device(max_idx))
        for idx in np.ix_(*tuple(np.arange(cc.shape[i]) for i in range(cc.ndim - 1)))
    )
    maxi = cc[(*indices, max_idx)]
    shift = unravel_index(max_idx, spatial_shape)
    return maxi**2, shift, z


def _phase_cross_correlation_broadcasted_array_api(
    reference_image: "Array",
    moving_image: "Array",
    *,
    upsample_factor: int = 1,
    space: str = "real",
    disambiguate: bool = False,
    reference_mask: Optional["Array"] = None,
    moving_mask: Optional["Array"] = None,
    overlap_ratio: float = 0.3,
    normalization: str = "phase",
    nb_spatial_dims: Optional[int] = None,
    multichannel: bool = False,
) -> Tuple["Array", "Array", "Array"]:
    """Phase cross-correlation between a reference and moving_images
    Params:
        reference (Array): image of shape ({...}, (C), [...])
            where [...] corresponds to the N spatial dimensions
        moving_images (Array): images to register of shape ({{...}}, (C), [...])
            where [...] corresponds to the N spatial dimensions
        upsample_factor (float): upsampling factor.
            Images will be registered up to 1/upsample_factor.
        space: not implemented
        disambiguate: not implemented
        reference_mask: not implemented
        moving_mask: not implemented
        overlap_ratio: not implemented
        normalization : {"phase", None}
            The type of normalization to apply to the cross-correlation. This
            parameter is unused when masks (`reference_mask` and `moving_mask`) are
            supplied.
        nb_spatial_dims (int): specify the N spatial dimensions
        multichannel (bool): optional C dimension.
    Returns:
        {...} and {{...}} shapes are broadcasted to (...)
        error (Array): tensor of shape (...)
        shift (Tuple[Array]): tuple of N tensors of size (...)
    """
    xp = array_namespace(reference_image, moving_image)
    device1, device2 = get_device(reference_image), get_device(moving_image)
    if device1 != device2:
        raise ValueError(f"found args on two devices: {device1}, {device2}")
    device = device1
    if disambiguate:
        raise NotImplementedError(
            "array api masked cross correlation disambiguate is not implemented"
        )
    if reference_mask is not None or moving_mask is not None:
        raise NotImplementedError(
            "array api masked cross correlation is not implemented"
        )
    if nb_spatial_dims is None:
        assert reference_image.shape == moving_image.shape
        nb_spatial_dims = reference_image.ndim
        if multichannel:
            nb_spatial_dims -= 1
    if not multichannel:
        reference_image = xp.expand_dims(reference_image, axis=-nb_spatial_dims - 1)
        moving_image = xp.expand_dims(moving_image, axis=-nb_spatial_dims - 1)
    output_shape = np.broadcast_shapes(reference_image.shape, moving_image.shape)
    if space == "real":
        (floating_dtype,) = set((reference_image.dtype, moving_image.dtype))
        assert floating_dtype in (xp.float32, xp.float64)
        if floating_dtype == xp.float32:
            complex_dtype = xp.complex64
        else:
            complex_dtype = xp.complex128
        reference_image_freq = xp.fft.fftn(
            xp.astype(reference_image, complex_dtype),
            axes=tuple(range(-nb_spatial_dims, 0)),
        )
        moving_image_freq = xp.fft.fftn(
            xp.astype(moving_image, complex_dtype),
            axes=tuple(range(-nb_spatial_dims, 0)),
        )
    elif space == "fourier":
        (complex_dtype,) = set((reference_image.dtype, moving_image.dtype))
        reference_image_freq = xp.astype(reference_image, complex_dtype)
        moving_image_freq = xp.astype(moving_image, complex_dtype)
    other_shapes, spatial_shapes = (
        output_shape[: -nb_spatial_dims - 1],
        output_shape[-nb_spatial_dims:],
    )
    channel_shape = (output_shape[-nb_spatial_dims - 1],)
    assert other_shapes + channel_shape + spatial_shapes == output_shape
    assert len(spatial_shapes) == nb_spatial_dims
    midpoints = xp.asarray(
        [axis_size // 2 for axis_size in spatial_shapes], device=device, dtype=xp.int64
    )

    # Single pixel registration
    error, shift, image_product = _cross_correlation_multichannel_max(
        reference_image_freq,
        moving_image_freq,
        normalization,
        nb_spatial_dims=nb_spatial_dims,
    )

    # Now change shifts so that they represent relative shifts and not indices
    spatial_shapes_broadcasted = xp.asarray(
        np.broadcast_to(spatial_shapes, shift.shape), device=device
    )
    shift[shift > midpoints] -= spatial_shapes_broadcasted[shift > midpoints]

    rg00 = xp.sum(
        (reference_image_freq * xp.conj(reference_image_freq)),
        axis=tuple(
            range(
                reference_image_freq.ndim - nb_spatial_dims - 1,
                reference_image_freq.ndim,
            )
        ),
    )
    rf00 = xp.sum(
        (moving_image_freq * xp.conj(moving_image_freq)),
        axis=tuple(
            range(moving_image_freq.ndim - nb_spatial_dims - 1, moving_image_freq.ndim)
        ),
    )

    if upsample_factor == 1:
        spatial_size = xp.prod(
            xp.asarray(spatial_shapes, dtype=complex_dtype, device=device)
        )
        rg00 = rg00 / spatial_size
        rf00 = rf00 / spatial_size
    else:
        shift = xp.astype(shift, xp.float32)
        shift = xp.round(shift * upsample_factor) / upsample_factor
        upsampled_region_size = math.ceil(upsample_factor * 1.5)
        dftshift = math.trunc(upsampled_region_size / 2.0)
        sample_region_offset = dftshift - shift * upsample_factor
        cross_correlation = xp.conj(
            _upsampled_dft(
                xp.conj(image_product),
                upsampled_region_size,
                upsample_factor,
                sample_region_offset,
                nb_spatial_dims,
            )
        )
        cross_correlation = xp.abs(cross_correlation)
        cross_correlation_reshaped = xp.reshape(
            cross_correlation, (*tuple(other_shapes), -1)
        )
        max_idx = xp.argmax(cross_correlation_reshaped, axis=-1)
        indices = tuple(
            xp.asarray(idx, dtype=max_idx.dtype, device=device)
            for idx in np.ix_(
                *tuple(
                    np.arange(cross_correlation_reshaped.shape[i])
                    for i in range(cross_correlation_reshaped.ndim - 1)
                )
            )
        )
        error = cross_correlation_reshaped[(*indices, max_idx)] ** 2
        maxima = unravel_index(max_idx, cross_correlation.shape[-nb_spatial_dims:])
        maxima -= dftshift

        shift += xp.astype(maxima, xp.float32) / xp.asarray(
            upsample_factor, dtype=xp.float32, device=device
        )

    error = 1.0 - error / (xp.real(rg00) * xp.real(rf00))
    error = xp.sqrt(xp.abs(error))

    return tuple([shift[..., i] for i in range(shift.shape[-1])]), error, None


def cartesian_prod(*arrays: "Array"):
    xp = array_namespace(*arrays)
    return xp.reshape(
        xp.stack(xp.meshgrid(*arrays, indexing="ij"), axis=-1), (-1, len(arrays))
    )


def discretize_sphere_uniformly(
    xp: "array_api_module",
    N: int,
    M: int,
    symmetry: int = 1,
    product: bool = False,
    dtype=None,
    device=None,
):
    """Generates a list of the two first euler angles that describe a uniform
    discretization of the sphere with the Fibonnaci sphere algorithm.
    Params:
        xp: numpy, torch or cupy
        N, the number of axes (two first euler angles)
        M, the number of rotations around the axes (third euler angle)
            symmetry, the order of symmetry to reduce the range of the 3rd angle.
            Default to 1, no symmetry product
            If True return the cartesian product between the axes and the rotations

    Returns: (theta, phi, psi), precision
        precision, a float representing an approximation of the sampling done
        (theta, phi, psi), a tuple of 1D arrays containing the 3 euler angles
            theta.shape == phi.shape == (N,)
            psi.shape == (M,)
        if product is true,
            theta.shape == phi.shape == psi.shape == (N*M,)
    """
    epsilon = 0.5
    goldenRatio = (1 + 5**0.5) / 2
    i = xp.arange(0, N, device=device, dtype=dtype)
    theta = xp.remainder(
        2 * xp.pi * i / goldenRatio, xp.asarray(2 * xp.pi, device=device, dtype=dtype)
    )
    phi = xp.acos(1 - 2 * (i + epsilon) / N)
    psi = xp.linspace(0, 2 * np.pi / symmetry, M, device=device, dtype=dtype)
    if product:
        theta, psi2 = cartesian_prod(theta, psi).T
        phi, _ = cartesian_prod(phi, psi).T
        psi = psi2

    precision_axes: float = (
        (180 / xp.pi) * 2 * xp.pi**0.5 / N**0.5
    )  # aire autour d'un point = 4*pi/N
    precision_rot = (180 / xp.pi) * 2 * xp.pi / symmetry / M
    theta, phi, psi = theta * 180 / xp.pi, phi * 180 / xp.pi, psi * 180 / xp.pi
    return (theta, phi, psi), (precision_axes, precision_rot)


def center_of_mass(volume: "Array"):
    xp = array_namespace(volume)
    if not xp.isdtype(volume.dtype, "real floating"):
        dtype = xp.float64
        assert not xp.isdtype(volume.dtype, "complex floating")
        volume_float = xp.astype(volume, xp.float64)
    else:
        dtype = volume.dtype
        volume_float = volume
    kwargs = dict(dtype=dtype, device=get_device(volume))
    zz, yy, xx = xp.meshgrid(
        xp.arange(volume.shape[-3], **kwargs),
        xp.arange(volume.shape[-2], **kwargs),
        xp.arange(volume.shape[-1], **kwargs),
        indexing="ij",
    )
    S = xp.sum(volume_float)
    return (
        float(xp.sum(volume_float * zz) / S),
        float(xp.sum(volume_float * yy) / S),
        float(xp.sum(volume_float * xx) / S),
    )


def translate(volume: "Array", vec: "Array", order: int = 1):
    """Translate volume of vec
    Args:
        volume: shape (D, H, W)
        vec: shape (3,)
    """
    xp = array_namespace(volume, vec)
    (device,) = set((get_device(volume), get_device(vec)))
    return affine_transform(
        volume,
        xp.eye(3, device=device, dtype=vec.dtype),
        offset=-vec,
        output_shape=volume.shape,
        order=order,
    )


def move_center_of_mass_to_center(volume: "Array", order: int = 1):
    xp = array_namespace(volume)
    tvec = (xp.asarray(volume.shape) - 1) / 2 - xp.asarray(center_of_mass(volume))
    return translate(volume, xp.asarray(tvec, device=get_device(volume)), order=order)


def disp3D(*ims: np.ndarray, fig=None, axis_off=False):
    if fig is None:
        fig = plt.figure()
    axes = fig.subplots(1, len(ims))
    if len(ims) == 1:
        axes = [axes]
    for i in range(len(ims)):
        views = [
            ims[i][ims[i].shape[0] // 2, :, :],
            ims[i][:, ims[i].shape[1] // 2, :],
            ims[i][:, :, ims[i].shape[2] // 2],
        ]
        axes[i].set_aspect(1.0)
        # views = [normalize_patches(torch.from_numpy(v)).cpu().numpy() for v in views]

        divider = make_axes_locatable(axes[i])
        # below height and pad are in inches

        ax_x = divider.append_axes(
            "right",
            size=1,
            pad=0,
            sharey=axes[i],
        )
        ax_y = divider.append_axes(
            "bottom",
            size=1,
            pad=0,
            sharex=axes[i],
        )

        # make some labels invisible
        axes[i].xaxis.set_tick_params(
            labeltop=True, top=True, labelbottom=False, bottom=False
        )
        ax_x.yaxis.set_tick_params(labelleft=False, left=False, right=True)
        ax_x.xaxis.set_tick_params(
            top=True, labeltop=True, bottom=True, labelbottom=False
        )
        ax_y.xaxis.set_tick_params(bottom=True, labelbottom=False, top=False)
        ax_y.yaxis.set_tick_params(right=True)

        # show slice info
        if not axis_off:
            axes[i].text(
                0,
                2,
                f"Z={ims[i].shape[0]//2}",
                color="white",
                bbox=dict(boxstyle="square"),
            )
            ax_x.text(
                0,
                2,
                f"Y={ims[i].shape[1]//2}",
                color="white",
                bbox=dict(boxstyle="square"),
            )
            ax_y.text(
                0,
                2,
                f"X={ims[i].shape[2]//2}",
                color="white",
                bbox=dict(boxstyle="square"),
            )
        vmin = min(views[0].min(), views[1].min(), views[2].min())
        vmax = max(views[0].max(), views[1].max(), views[2].max())
        axes[i].imshow(views[0], cmap="gray", vmin=vmin, vmax=vmax)
        ax_y.imshow(views[1], cmap="gray", vmin=vmin, vmax=vmax)
        ax_x.imshow(ndii.rotate(views[2], 90)[::-1], cmap="gray", vmin=vmin, vmax=vmax)
        axes[i].set_xlim(0, ims[i].shape[2] - 1)
        axes[i].set_ylim(0, ims[i].shape[1] - 1)
        axes[i].yaxis.set_inverted(True)

    if axis_off:
        for ax in axes:
            ax.set_axis_off()
    return fig, axes


def disp2D(*ims, fig=None, **imshowkwargs):
    if fig is None:
        fig = plt.figure()
    h = int(math.floor(len(ims) ** 0.5))
    w = int(math.ceil(len(ims) / h))
    axes = fig.subplots(h, w)
    if isinstance(axes, np.ndarray):
        axes = axes.flatten()
        for ax in axes:
            ax.set_axis_off()
        for i in range(len(ims)):
            axes[i].imshow(ims[i], **imshowkwargs)
    else:
        axes.set_axis_off()
        axes.imshow(ims[0], **imshowkwargs)
    return fig, axes


def disp2D_compare(fig, *ims, **imshowkwargs):
    h = int(np.floor(len(ims) ** 0.5))
    w = int(np.ceil(len(ims) / h))
    axes = fig.subplots(h, w)
    if isinstance(axes, np.ndarray):
        axes = axes.flatten()
        for ax in axes:
            ax.set_axis_off()
        for i in range(len(ims)):
            im = np.concatenate(tuple(ims[i]), axis=1)
            axes[i].imshow(im, **imshowkwargs)
    else:
        axes.set_axis_off()
        im = np.concatenate(tuple(ims[0]), axis=1)
        axes.imshow(im, **imshowkwargs)


def get_random_3d_vector(norm=None):
    """
    Generates a random 3D unit vector (direction) with a uniform spherical distribution
    Algo from http://stackoverflow.com/questions/5408276/python-uniform-spherical-distribution
    :return:
    """
    phi = np.random.uniform(0, np.pi * 2)
    costheta = np.random.uniform(-1, 1)

    theta = np.arccos(costheta)
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)
    if norm is None:
        norm = 1
    if norm < 0:
        norm = 0
    return norm * np.array([x, y, z])


def get_surfaces(self, corners_points):
    p = corners_points
    return np.array(
        [
            [p[0], p[1], p[3]],
            [p[0], p[4], p[3]],
            [p[0], p[4], p[1]],
            [p[6], p[7], p[2]],
            [p[6], p[2], p[5]],
            [p[6], p[7], p[5]],
        ]
    )


def get_plane_equation(self, s):
    normal1 = np.cross(s[1] - s[0], s[2] - s[0])
    normal1 /= np.linalg.norm(normal1)
    return np.concatenate([normal1, [-np.dot(normal1, s[0])]])


def get_planes_intersection(self, s1, s2):
    """tested"""
    equation1 = self.get_plane_equation(s1)
    equation2 = self.get_plane_equation(s2)

    vec1, vec2 = equation1[:3], equation2[:3]
    line_vec = np.cross(vec1, vec2)
    A = np.array([vec1, vec2, line_vec])
    d = np.array([-equation1[3], -equation2[3], 0.0]).reshape(3, 1)

    if np.linalg.det(A) == 0:
        return False, None
    else:
        p_inter = np.linalg.solve(A, d).T
        return True, (line_vec, p_inter[0])


def get_lines_intersection(self, eq1, eq2):
    A = np.array([eq1[:2], eq2[:2]])
    d = -np.array([eq1[2], eq2[2]])
    if np.linalg.det(A) == 0:
        return False, None
    else:
        p_inter = np.linalg.solve(A, d).T
        return True, p_inter


def line_crossing_segment(self, line, segment):
    """tested"""

    vec_line, p_line = line
    vec_segment = segment[1] - segment[0]
    A = np.array([vec_segment, -vec_line]).T
    d = (p_line - segment[0]).reshape(2, 1)
    if np.linalg.det(A) == 0:
        return False, None
    t1, t2 = np.linalg.solve(A, d).reshape(-1)
    return t1 >= 0 and t1 <= 1, t2


def line_intersect_surface(self, line, surface):
    """tested"""
    plane_basis = np.array([surface[1] - surface[0], surface[2] - surface[0]])
    plane_basis_position = surface[0]
    plane_orthonormal_basis = plane_basis / np.linalg.norm(plane_basis, axis=1)[:, None]

    def projector(x):
        return np.array(
            [
                np.dot(plane_orthonormal_basis[0], x - plane_basis_position),
                np.dot(plane_orthonormal_basis[1], x - plane_basis_position),
            ]
        )

    projected_line = (
        projector(line[1] + 10 * line[0]) - projector(line[1]),
        projector(line[1]),
    )
    projected_surface = np.array(
        [projector(surface[0]), projector(surface[1]), projector(surface[2])]
    )
    p4 = projected_surface[1] + projected_surface[2]
    segments = np.array(
        [
            [projected_surface[0], projected_surface[1]],
            [projected_surface[0], projected_surface[2]],
            [projected_surface[1], p4],
            [projected_surface[2], p4],
        ]
    )

    out = [self.line_crossing_segment(projected_line, seg) for seg in segments]
    t = list(map(lambda x: x[1], filter(lambda x: x[0], out)))
    if len(t) > 0:
        return True, (min(t), max(t))
    else:
        return False, (None, None)


def surfaces_intersect(self, s1, s2):
    ret, line = self.get_planes_intersection(s1, s2)
    if not ret:
        return False
    ret1, (tmin1, tmax1) = self.line_intersect_surface(line, s1)
    ret2, (tmin2, tmax2) = self.line_intersect_surface(line, s2)
    if ret1 and ret2:
        return ((tmin1 <= tmax2) and (tmin2 <= tmin1)) or (
            (tmin2 <= tmax1) and (tmin1 <= tmin2)
        )
    else:
        return False


def pointcloud_intersect(self, corners1, corners2):
    surfaces1 = self.get_surfaces(corners1)
    surfaces2 = self.get_surfaces(corners2)
    intersections = [
        self.surfaces_intersect(s1, s2)
        for s1, s2 in itertools.product(surfaces1, surfaces2)
    ]
    return any(intersections)


def are_volumes_aligned(vol1, vol2, atol=0.1, nb_spatial_dims=3):
    (dz, dy, dx), _, _ = phase_cross_correlation(
        vol1,
        vol2,
        upsample_factor=10,
        disambiguate=True,
        normalization=None,
        nb_spatial_dims=nb_spatial_dims,
    )
    n = (dz**2 + dy**2 + dx**2) ** 0.5
    return n <= atol


def assert_volumes_translated(vol1, vol2, atol=0.1, nb_spatial_dims=3):
    """check if vol1 is a translation of vol2"""
    _, error, _ = phase_cross_correlation(
        vol1,
        vol2,
        upsample_factor=10,
        disambiguate=True,
        normalization=None,
        nb_spatial_dims=nb_spatial_dims,
    )
    np.testing.assert_allclose(to_numpy(error), 0, atol=atol)


def tukey(
    xp: "array_api_module", shape: tuple[int], alpha: float = 0.5, sym: bool = True
):
    tukeys = [tukey_scipy(s, alpha=alpha, sym=sym) for s in shape]
    tukeys_reshaped = [
        xp.reshape(t, (1,) * i + (-1,) + (1,) * (len(shape) - i - 1))
        for i, t in enumerate(tukeys)
    ]
    final_window = tukeys_reshaped[0]
    for t in tukeys_reshaped[1:]:
        final_window = final_window * t
    return xp.asarray(final_window)



from __future__ import annotations

from typing import TYPE_CHECKING

from skimage.filters import threshold_otsu as skimage_threshold_otsu
from sklearn.cluster import KMeans
from sklearn.svm import SVC

from spfluo.utils.array import array_namespace, numpy_only_compatibility
from spfluo.utils.rotate_symmetry_axis import find_rotation_between_two_vectors
from spfluo.utils.volume import affine_transform, tukey

if TYPE_CHECKING:
    from spfluo.utils.array import Array


@numpy_only_compatibility
def _threshold_otsu(image: "Array", nbins: int = 256):
    return skimage_threshold_otsu(image, nbins=nbins)


def extract_particle(
    image_data: "Array",
    pos: tuple[float, float, float],
    dim: tuple[float, float, float],
    scale: tuple[float, float, float],
    subpixel: bool = False,
):
    xp = array_namespace(image_data)

    def world_to_data_coord(pos):
        return pos / xp.asarray(scale)

    def data_to_world_coords(pos):
        return pos * xp.asarray(scale)

    pos = world_to_data_coord(xp.asarray(pos))  # World coordinates to data coords
    box_size_world = xp.asarray(dim, dtype=float)
    box_size_data = xp.astype(xp.ceil(world_to_data_coord(box_size_world)), xp.int64)
    real_dim = data_to_world_coords(box_size_data)
    mat = xp.eye(4)
    mat[:3, 3] = pos
    mat[:3, 3] -= box_size_data / 2
    C = image_data.shape[0]
    particle_data = xp.empty((C,) + tuple(box_size_data), dtype=image_data.dtype)
    if not subpixel:
        top_left_corner = xp.round(mat[:3, 3]).astype(int)
        bottom_right_corner = top_left_corner + box_size_data
        xmin, ymin, zmin = top_left_corner
        xmax, ymax, zmax = bottom_right_corner
        original_shape = image_data.shape[1:]
        x_slice = slice(max(xmin, 0), min(xmax, original_shape[0]))
        y_slice = slice(max(ymin, 0), min(ymax, original_shape[1]))
        z_slice = slice(max(zmin, 0), min(zmax, original_shape[2]))
        x_overlap = slice(
            max(0, -xmin), min(box_size_data[0], original_shape[0] - xmin)
        )
        y_overlap = slice(
            max(0, -ymin), min(box_size_data[1], original_shape[1] - ymin)
        )
        z_overlap = slice(
            max(0, -zmin), min(box_size_data[2], original_shape[2] - zmin)
        )

    for c in range(C):
        if subpixel:
            particle_data[c] = affine_transform(
                image_data[c], mat, output_shape=tuple(box_size_data)
            )
        else:
            padded_array = xp.zeros(tuple(box_size_data), dtype=image_data.dtype)
            padded_array[x_overlap, y_overlap, z_overlap] = image_data[
                c, x_slice, y_slice, z_slice
            ]
            particle_data[c] = xp.asarray(padded_array, copy=True)

    return particle_data, real_dim


def separate_centrioles(
    im: "Array",
    output_size: tuple[int, int, int],
    threshold_percentage: float = 0.5,
    channel: int = 0,
    tukey_alpha: float = 0.1,
):
    xp = array_namespace(im)
    return separate_centrioles_coords(
        im,
        (xp.asarray(im.shape[-3:]) - 1) / 2,
        im.shape[-3:],
        output_size,
        threshold_percentage=threshold_percentage,
        channel=channel,
        tukey_alpha=tukey_alpha,
    )


def kmeans_cluster(points):
    # Clustering
    kmeans = KMeans(n_clusters=2, n_init="auto")
    kmeans.fit(points)

    return kmeans.labels_, kmeans.cluster_centers_


def separate_centrioles_coords(
    image: "Array",
    pos: tuple[float, float, float],
    dim: tuple[float, float, float],
    output_size: tuple[float, float, float],
    *,
    scale: tuple[float, float, float] = (1, 1, 1),
    threshold_percentage: float = 0.5,
    channel: int = 0,
    tukey_alpha: float = 0.1,
):
    xp = array_namespace(image)

    if image.ndim > 3:
        multichannel = True
    else:
        multichannel = False
        image = image[None]

    # extract patch from image
    patch = extract_particle(image, pos, dim, scale, subpixel=False)[0][channel]

    # Thresholding
    mask = patch > xp.max(patch) * threshold_percentage
    points = xp.stack(xp.nonzero(mask), axis=-1)

    # Cluster in the real space
    patch_top_left_corner = xp.asarray(pos) - xp.asarray(dim) / 2
    points = points * scale + patch_top_left_corner  # points in the image space
    labels, _ = kmeans_cluster(points)

    (patch1, patch2), _ = _separate_clusters(
        image, points, labels, patch[mask], output_size, scale, tukey_alpha
    )

    if not multichannel:
        patch1, patch2 = patch1[0], patch2[0]

    patch1, patch2 = (
        xp.asarray(patch1, dtype=image.dtype),
        xp.asarray(patch2, dtype=image.dtype),
    )
    return patch1, patch2


def _run_svc(clusters_points, clusters_labels, intensities):
    svc = SVC(kernel="linear")
    svc.fit(clusters_points, clusters_labels, intensities)
    return svc.coef_, svc.intercept_


def _separate_clusters(
    image: "Array",
    clusters_points: "Array",
    clusters_labels: "Array",
    cluster_intensities: "Array",
    output_size: tuple[float, ...],
    scale: tuple[float, ...],
    tukey_alpha: float,
):
    """Separate image into 2 images
    Arguments:
        - image: full image of shape (C, D, H, W)
        - clusters_points: points of the clusters
    """
    xp = array_namespace(image, clusters_points, clusters_labels)
    assert clusters_labels.max() == 1
    centers = xp.stack(
        (
            xp.mean(clusters_points[clusters_labels == 0], axis=0),
            xp.mean(clusters_points[clusters_labels == 1], axis=0),
        ),
        axis=0,
    )

    # extract 2 patches around the centroids
    image_float = xp.asarray(image, dtype=xp.float64)
    patch1, real_dim = extract_particle(
        image_float, centers[0], output_size, scale, subpixel=False
    )
    patch2, _ = extract_particle(
        image_float, centers[1], output_size, scale, subpixel=False
    )

    plane_coef, plane_intercept = _run_svc(
        clusters_points, clusters_labels, cluster_intensities
    )

    def proj(x):
        return xp.vecdot(x, xp.asarray(plane_coef)) + xp.asarray(plane_intercept)

    # Compute hyperplane for each patch
    s = xp.max(patch1.shape[-3:])
    size_tukey = xp.asarray([s, s, s]) * 3**0.5
    t = tukey(xp, xp.asarray(xp.round(size_tukey), dtype=xp.int64), alpha=tukey_alpha)

    # Compute hyperplane for each patch and apply tukey window
    for patch, pos_patch in ((patch1, centers[0]), (patch2, centers[1])):
        image_coords = xp.stack(
            xp.meshgrid(*[xp.arange(s) for s in patch.shape[1:]], indexing="ij"),
            axis=-1,
        )
        patch_top_left_corner = xp.asarray(pos_patch) - xp.asarray(output_size) / 2
        image_coords = image_coords * scale + patch_top_left_corner  # to image space
        pos_patch_proj = proj(pos_patch)

        # Compute the tukey window
        v = xp.asarray(plane_coef[0])
        d = xp.asarray(plane_intercept) / xp.linalg.norm(v)
        v = v / xp.linalg.norm(v)
        unit = xp.asarray([0, 1, 0])
        R = find_rotation_between_two_vectors(unit, v)

        H1 = xp.eye(4)
        H1[:3, 3] = -xp.asarray(scale) * xp.round(size_tukey) / 2
        H1[[0, 1, 2], [0, 1, 2]] = xp.asarray(scale)  # pixel to world space
        H2 = xp.eye(4)  # world space
        H2[:3, :3] = R  # world space
        pos_tukey_plane = (
            R
            @ (
                -xp.sign(pos_patch_proj)
                * xp.asarray(scale)
                * unit
                * xp.round(size_tukey)
                / 2
            )
            / xp.asarray(scale)
        )  # pixel space

        center_patch = xp.asarray(pos_patch)  # world space
        patch_top_left_corner = (
            xp.asarray(center_patch) - xp.asarray(output_size) / 2
        )  # world space
        center_patch_proj = center_patch - v * (
            xp.vecdot(center_patch, v) + d
        )  # world space
        center_patch_proj_patch_space = (
            center_patch_proj - patch_top_left_corner
        ) / xp.asarray(scale)  # pixel space
        H3 = xp.eye(4)
        H3[[0, 1, 2], [0, 1, 2]] = 1 / xp.asarray(scale)
        H3[:3, 3] = center_patch_proj_patch_space - pos_tukey_plane
        H = H3 @ H2 @ H1
        t_rotated = affine_transform(t, xp.linalg.inv(H), output_shape=patch.shape[-3:])

        patch *= t_rotated

    # Contrast enhancement
    for patch in (patch1, patch2):
        otsu = _threshold_otsu(patch)
        patch *= (xp.tanh((patch - otsu) / (otsu * 0.1)) + 1) / 2

    return (patch1, patch2), centers



from pathlib import Path

from spfluo.data.upload import _file_hash


def make_registry(directory: Path, output: Path, extensions: list[str] = ["*"]):
    """taken from pooch"""

    files = sorted(
        str(path.relative_to(directory))
        for ext in extensions
        for path in directory.glob(f"**/*{ext}")
        if path.is_file()
    )

    hashes = [_file_hash(directory / fname) for fname in files]

    with open(output, "w") as outfile:
        for fname, fhash in zip(files, hashes):
            # Only use Unix separators for the registry so that we don't go
            # insane dealing with file paths.
            outfile.write("{} {}\n".format(fname.replace("\\", "/"), fhash))


if __name__ == "__main__":
    data_dir = Path(__file__).parent
    make_registry(data_dir, data_dir / "registry.txt", extensions=[".csv", ".tiff", ".tif"])



import os
import pickle
from typing import Dict, List, Tuple

import numpy as np
import torch
from tqdm import tqdm

from ..training import EfficientNet, make_U_mask
from ..training.prepare_fs_data import get_patch_paths
from ..utils import load_array, summary

# +------------------------------------------------------------------------------------------+ #
# |                                          HELPERS                                         | #
# +------------------------------------------------------------------------------------------+ #


def load_network(
    network: str, dim: int, num_classes: int, checkpoint_path: str, device: torch.device
) -> torch.nn.Module:
    network = EfficientNet.from_name(network, dim=dim, num_classes=num_classes).to(
        device
    )
    checkpoint = torch.load(checkpoint_path)
    network.load_state_dict(checkpoint["network_state_dict"])
    network = network.to(device)
    network = network.eval()
    return network


def make_patches(
    image: torch.Tensor, patch_size: Tuple[int], stride: Tuple[int], batch_size: int = 1
):
    pz, py, px = patch_size
    sz, sy, sx = stride

    image_batch_buffer = torch.empty((batch_size,) + patch_size, dtype=image.dtype)
    coords_batch_buffer = torch.empty((batch_size,) + (3,), dtype=torch.int64)
    i = 0
    for z in range(0, image.shape[0] - pz + 1, sz):
        for y in range(0, image.shape[1] - py + 1, sy):
            for x in range(0, image.shape[2] - px + 1, sx):
                image_batch_buffer[i] = image[z : z + pz, y : y + py, x : x + px]
                coords_batch_buffer[i] = torch.as_tensor(
                    (z + pz // 2, y + py // 2, x + px // 2), dtype=torch.int64
                )
                i += 1
                if i == batch_size:
                    yield (
                        torch.clone(image_batch_buffer),
                        torch.clone(coords_batch_buffer),
                    )
                    i = 0
    if i > 0:
        yield torch.clone(image_batch_buffer[:i]), torch.clone(coords_batch_buffer)


def flatten_patch_index_to_image_coord(
    index: int, unfold_shape: torch.Size, stride: Tuple[int]
) -> Tuple[int]:
    Z, Y, X, pz, py, px = unfold_shape
    sz, sy, sx = stride
    z = index // (Y * X)
    yx = index % (Y * X)
    y = yx // X
    x = yx % X
    return z * sz + pz // 2, y * sy + py // 2, x * sx + px // 2


# +------------------------------------------------------------------------------------------+ #
# |                                  MAIN PREDICTION FUNCTION                                | #
# +------------------------------------------------------------------------------------------+ #


@torch.no_grad()
def predict_one_image_picking(
    image: torch.Tensor,
    patch_size: Tuple[int],
    dim: int,
    network: torch.nn.Module,
    device: torch.device,
    batch_size: int,
    stride: Tuple[int] = None,
    predict_on_u_mask: bool = False,
    progress_bar: bool = True,
) -> Tuple[np.ndarray]:
    patches_generator = make_patches(image, patch_size, stride, batch_size=batch_size)
    if predict_on_u_mask:
        # make_u_mask works on numpy array so some casting overhead is required.
        u_mask = make_U_mask(image.numpy())
        u_mask = torch.as_tensor(u_mask, dtype=torch.float32)
        u_mask_patches_generator = make_patches(
            u_mask, patch_size, stride, batch_size=batch_size
        )
    else:
        u_mask_patches_generator = (None for _ in iter(int, 1))  # infinite None
    coords: list[tuple[int, int, int]] = []
    scores: list[float] = []
    for (patches, patches_coords), (u_mask_patches, _) in zip(
        patches_generator, u_mask_patches_generator
    ):
        # Filter out patches not on U masks
        if u_mask_patches is not None:
            pz, py, px = patch_size
            filtered_indices = torch.where(
                u_mask_patches[:, pz // 2, py // 2, px // 2] > 0
            )[0]
        else:
            filtered_indices = torch.arange((patches.size(0),), dtype=torch.int64)
        patches = patches[filtered_indices]
        patches_coords = patches_coords[filtered_indices]

        if patches.size(0) > 0:
            patches = torch.stack(
                [(p - p.min()) / (p.max() - p.min()) for p in patches]
            )
            patches = patches.unsqueeze(1)  # add channels dim

            # Project
            if dim == 2:
                patches = patches.sum(axis=-3)

            # Predict
            inputs = patches.to(device)
            outputs = network(inputs)
            positive_indices = torch.where(outputs > 0)[0]
            for batch_index in positive_indices:
                scores.append(outputs[batch_index].item())
                coords.append(patches_coords[batch_index].tolist())
    return np.array(coords), np.array(scores)


@torch.no_grad()
def predict_patches_tilt(
    image: str,
    patches_path: List[str],
    rootdir: str,
    extension: str,
    dim: int,
    network: torch.nn.Module,
    device: torch.device,
    batch_size: int,
    progress_bar: bool = True,
) -> Tuple[np.ndarray]:
    test_paths_image = list(
        filter(
            lambda x: "_".join(os.path.basename(x).split("_")[:-1]) + "." + extension
            == image,
            patches_path,
        )
    )
    test_paths_image.sort(
        key=lambda x: int(
            os.path.basename(x).split("_")[-1].split(".")[0],
        )
    )
    patch_names = list(map(lambda x: os.path.basename(x), test_paths_image))
    patches = []
    for crop_path in test_paths_image:
        patch_path = os.path.join(rootdir, crop_path)
        arr = load_array(patch_path)
        if arr.dtype == np.uint16:
            arr = arr.astype(np.uint8)
        patch = torch.as_tensor(arr, dtype=torch.float32)
        patches.append(patch)
    patches = torch.stack(patches, dim=0)

    if dim == 2:
        patches = patches.sum(axis=1)
    flatten_patches = patches.view(patches.size(0), -1)
    (min_patch, _), (max_patch, _) = (
        flatten_patches.min(dim=1),
        flatten_patches.max(dim=1),
    )
    min_patch = min_patch.view(-1, 1, 1)
    max_patch = max_patch.view(-1, 1, 1)
    patches = (patches - min_patch) / (max_patch - min_patch)
    patches = patches.unsqueeze(1)

    num_batches = patches.size(0) // batch_size
    generator = tqdm(range(num_batches)) if progress_bar else range(num_batches)
    scores, labels = [], []
    for i in generator:
        inputs = patches[i * batch_size : (i + 1) * batch_size].to(device)
        outputs = network(inputs)
        for batch_index in range(len(outputs)):
            score = outputs[batch_index].cpu().numpy()
            scores.append(score)
            label = score.argmax()
            labels.append(label)

    if patches.size(0) % batch_size != 0:
        inputs = patches[num_batches * batch_size :].to(device)
        outputs = network(inputs)
        for batch_index in range(len(outputs)):
            score = outputs[batch_index].cpu().numpy()
            scores.append(score)
            label = score.argmax()
            labels.append(label)

    scores = np.stack(scores, axis=0)
    return np.array(labels), scores, np.array(patch_names)


# +------------------------------------------------------------------------------------------+ #
# |                                       EXTERNAL CALL                                      | #
# +------------------------------------------------------------------------------------------+ #


def predict_picking(
    network: str,
    checkpoint_path: str,
    rootdir: str,
    predict_on_u_mask: bool,
    patch_size: Tuple[int],
    dim: int,
    stride: Tuple[int],
    batch_size: int,
    extension: str,
    testset_size: int = None,
    image_name: str = None,
    output_dir: str = None,
    downscale: float = 1.0,
) -> Dict[str, Tuple[np.ndarray]]:
    """If image_name is None, predict whole folder."""
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    patch_size = tuple(np.rint(np.array(patch_size) / downscale).astype(int))
    stride = (
        np.array(patch_size) // 4
        if stride is None
        else tuple(np.rint(np.array(stride) / downscale).astype(int))
    )
    summary_output = None
    if output_dir is not None:
        os.makedirs(output_dir, exist_ok=True)
        summary_output = os.path.join(output_dir, "prediction_config.txt")
    summary(locals(), title="PREDICTION", output=summary_output)
    network = load_network(network, dim, 1, checkpoint_path, device)
    predict_args = [
        patch_size,
        dim,
        network,
        device,
        batch_size,
        stride,
        predict_on_u_mask,
    ]
    if image_name is not None:
        images_names = [image_name]
    else:
        images_names = sorted(
            list(filter(lambda x: x.endswith(extension), os.listdir(rootdir)))
        )
        images_names = images_names[:testset_size]
    predictions = {}
    for image_name in images_names:
        image_path = os.path.join(rootdir, image_name)
        image = torch.as_tensor(load_array(image_path), dtype=torch.float32)
        predictions[image_name] = predict_one_image_picking(image, *predict_args)
    if output_dir is not None:
        with open(os.path.join(output_dir, "predictions.pickle"), "wb") as file:
            pickle.dump(predictions, file)
    return predictions


def predict_tilt(
    rootdir: str,
    network: str,
    checkpoint_path: str,
    crop_dir: str,
    dim: int,
    batch_size: int,
    extension: str,
    testset_size: int = None,
    image_name: str = None,
    output_dir: str = None,
) -> Dict[str, Tuple[np.ndarray]]:
    """If image_name is None, predict whole folder."""
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    summary_output = None
    if output_dir is not None:
        summary_output = os.path.join(output_dir, "prediction_config.txt")
        os.makedirs(output_dir, exist_ok=True)
    summary(locals(), title="PREDICTION", output=summary_output)
    network = load_network(network, dim, 3, checkpoint_path, device)
    predict_args = [extension, dim, network, device, batch_size]

    if image_name is not None:
        images_names = [image_name]
    else:
        images_names = sorted(
            list(filter(lambda x: x.endswith(extension), os.listdir(rootdir)))
        )
        images_names = images_names[:testset_size]

    test_paths = get_patch_paths(crop_dir)
    predictions = {}
    for image_name in images_names:
        predictions[image_name] = predict_patches_tilt(
            image_name, test_paths, crop_dir, *predict_args
        )

    if output_dir is not None:
        with open(os.path.join(output_dir, "predictions.pickle"), "wb") as file:
            pickle.dump(predictions, file)

    return predictions



import typing

from spfluo.refinement.refinement import create_poses_grid, reconstruction_L2
from spfluo.utils.array import array_namespace
from spfluo.utils.transform import (
    compose_poses,
    get_transform_matrix_from_pose,
    symmetrize_poses,
)
from spfluo.utils.volume import affine_transform, phase_cross_correlation

if typing.TYPE_CHECKING:
    from spfluo.utils.array import Array


def reconstruct(
    centriole_top: "Array",
    centriole_side: "Array",
    pose_side: "Array",
    psf: "Array",
    /,
    lambda_=1.0,
    M=100,
):
    xp = array_namespace(centriole_top, centriole_side)
    pose_top = xp.asarray(
        [0, 0, 0, 0, 0, 0], dtype=xp.float64
    )  # top particle is the reference

    assert centriole_side.shape == centriole_side.shape == psf.shape
    assert centriole_side.ndim == 3 or centriole_side.ndim == 4

    multichannel = centriole_side.ndim == 4

    # Find translation between centriole side and top
    centriole_side_rotated_back = affine_transform(
        centriole_side,
        get_transform_matrix_from_pose(centriole_side.shape, pose_side),
        multichannel=multichannel,
    )
    shift, error, phasediff = phase_cross_correlation(
        centriole_top,
        centriole_side_rotated_back,
        upsample_factor=10,
        normalization=None,
        multichannel=multichannel,
    )
    pose_reference_to_centriole_side_rotated_back = xp.concat(
        (xp.asarray([0, 0, 0], dtype=xp.float64), -shift)
    )
    pose_side_registered = compose_poses(
        pose_reference_to_centriole_side_rotated_back, pose_side
    )

    poses_symmetry, _ = create_poses_grid(xp, 1, M, symmetry=9)
    poses_side_registered_symmetry = compose_poses(poses_symmetry, pose_side_registered)

    centriole_side_registered_sym = affine_transform(
        xp.stack((centriole_side,) * M),
        get_transform_matrix_from_pose(
            centriole_side.shape, poses_side_registered_symmetry
        ),
        batch=True,
        multichannel=multichannel,
    )

    _, errors, _ = phase_cross_correlation(
        centriole_top,
        centriole_side_registered_sym,
        upsample_factor=1,
        normalization=None,
        nb_spatial_dims=3,
        multichannel=multichannel,
    )
    best = xp.argmin(errors)
    best_pose_side = poses_side_registered_symmetry[best]

    # reconstruct
    poses = xp.stack(
        (symmetrize_poses(best_pose_side, 9), symmetrize_poses(pose_top, 9))
    )
    poses = xp.permute_dims(poses, (1, 0, 2))
    return reconstruction_L2(
        xp.stack((centriole_side, centriole_top)),
        psf,
        poses,
        lambda_=xp.asarray(lambda_),
        device="cpu",
        symmetry=True,
    )



import random

from ..augmentations import transforms as T


class Compose:
    def __init__(self, transforms, p=1.0, targets=[["image"], ["mask"]]):
        assert 0 <= p <= 1
        self.transforms = (
            [T.Float(always_apply=True)]
            + transforms
            + [T.Contiguous(always_apply=True)]
        )
        self.p = p
        self.targets = targets

    def get_always_apply_transforms(self):
        res = []
        for tr in self.transforms:
            if tr.always_apply:
                res.append(tr)
        return res

    def __call__(self, force_apply=False, **data):
        need_to_run = force_apply or random.random() < self.p
        transforms = (
            self.transforms if need_to_run else self.get_always_apply_transforms()
        )
        for tr in transforms:
            data = tr(force_apply, self.targets, **data)
        return data



from __future__ import annotations

import tifffile

from spfluo.ab_initio_reconstruction.common_image_processing_methods.others import (
    normalize,
)
from spfluo.refinement.refinement import reconstruction_L2
from spfluo.utils.array import array_namespace, get_device, to_numpy
from spfluo.utils.loading import read_poses
from spfluo.utils.read_save_files import read_image
from spfluo.utils.transform import symmetrize_poses


def main(
    particles_paths: list[str],
    poses_paths: str,
    psf_path: str,
    output_volume_path: str,
    lbda: float,
    symmetry: int = 1,
    gpu: bool = False,
    batch_size: int | None = None,
):
    assert output_volume_path.endswith(".ome.tiff")
    psf = normalize(read_image(psf_path, dtype="float64", gpu=gpu))
    xp = array_namespace(psf)
    compute_device = get_device(psf)
    psf = xp.to_device(psf, "cpu")
    host_device = get_device(psf)
    device = get_device(psf)
    particles = xp.stack(
        [
            normalize(read_image(p, dtype="float64", xp=xp, device=host_device))
            for p in particles_paths
        ]
    )
    if particles.ndim == 4:
        particles = particles[:, None]
    C = particles.shape[1]
    if psf.ndim == 3:
        psf = xp.stack((psf,) * C)
    poses, _ = read_poses(poses_paths, alphabetic_order=False)
    poses = xp.asarray(poses, dtype=xp.float64, device=host_device)
    poses = xp.permute_dims(symmetrize_poses(poses, symmetry), (1, 0, 2))
    reconstruction = xp.empty_like(particles[0])
    reconstruction = reconstruction_L2(
        particles,
        psf,
        poses,
        xp.asarray(lbda, dtype=particles.dtype, device=device),
        symmetry=True,
        device=compute_device,
        batch_size=batch_size,
        multichannel=True,
    )

    tifffile.imwrite(
        output_volume_path, to_numpy(reconstruction), metadata={"axes": "CZYX"}
    )



# from spfluo.utils.loading import loadmat
from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Tuple

import numpy as np
import pytest

import spfluo
from spfluo.refinement import (
    convolution_matching_poses_grid,
    convolution_matching_poses_refined,
    find_angles_grid,
    reconstruction_L2,
    refine,
)
from spfluo.tests.helpers import (
    assert_allclose,
    assert_volumes_aligned,
    ids,
    testing_libs,
)
from spfluo.utils.array import array_namespace, get_device, median
from spfluo.utils.transform import (
    distance_family_poses,
    get_transform_matrix,
    symmetrize_angles,
    symmetrize_poses,
)
from spfluo.utils.volume import (
    affine_transform,
)

if TYPE_CHECKING:
    from spfluo.utils.array import Array


@pytest.fixture(autouse=True)
def clear_cuda_cache():
    yield
    if spfluo.has_torch_cuda():
        import torch

        torch.cuda.empty_cache()
    if spfluo.has_cupy():
        import cupy

        mempool = cupy.get_default_memory_pool()
        mempool.free_all_blocks()


@pytest.fixture(scope="module")
def generated_data_all_array(request, generated_data_all):
    xp, device = request.param
    volumes, poses, psf, groundtruth = tuple(
        [xp.asarray(x) for x in generated_data_all]
    )
    return (volumes, poses, psf, groundtruth), device


@pytest.fixture
def poses_with_noise(
    generated_data_all_array: Tuple["Array", ...],
):
    (_, poses, _, _), _ = generated_data_all_array
    xp = array_namespace(poses)
    device = get_device(poses)
    poses_noisy = xp.asarray(poses, copy=True)
    sigma_rot, sigma_trans = 20, 2
    np.random.seed(0)
    poses_noisy[:, :3] += (
        xp.asarray(np.random.randn(len(poses), 3), device=device) * sigma_rot
    )
    poses_noisy[:, 3:] += (
        xp.asarray(np.random.randn(len(poses), 3), device=device) * sigma_trans
    )
    return poses_noisy


##########################
# Test reconstruction_L2 #
##########################


@pytest.mark.parametrize(
    "generated_data_all_array", testing_libs, indirect=True, ids=ids
)
def test_shapes_reconstruction_L2(generated_data_all_array):
    (volumes, groundtruth_poses, psf, groundtruth), device = generated_data_all_array
    xp = array_namespace(volumes)
    lambda_ = xp.asarray(1.0, device=get_device(volumes))
    recon = reconstruction_L2(volumes, psf, groundtruth_poses, lambda_, device=device)

    assert recon.shape == volumes.shape[-3:]


@pytest.mark.parametrize("minibatch", [1, 10, None])
@pytest.mark.parametrize(
    "generated_data_all_array", testing_libs, indirect=True, ids=ids
)
def test_parallel_reconstruction_L2(minibatch, generated_data_all_array, save_result):
    (volumes, groundtruth_poses, psf, groundtruth), device = generated_data_all_array
    M = 3
    xp = array_namespace(volumes)
    lambda_ = xp.ones((M,))
    poses = groundtruth_poses + xp.asarray(
        np.random.randn(M, volumes.shape[0], 6) * 0.1
    )
    recon = reconstruction_L2(
        volumes, psf, poses, lambda_, batch=True, device=device, batch_size=minibatch
    )
    recon2 = xp.stack(
        [
            reconstruction_L2(volumes, psf, poses[i, ...], lambda_[i], device=device)
            for i in range(M)
        ]
    )

    save_result("reconstructions", recon2, metadata={"axes": "TCZYX"})
    save_result("reconstructions_paralled", recon, metadata={"axes": "TCZYX"})

    assert recon.shape == (M,) + volumes.shape[-3:]
    assert_allclose(recon, recon2, atol=1e-5)


@pytest.mark.parametrize(
    "generated_data_all_array", testing_libs, indirect=True, ids=ids
)
def test_multichannel_reconstruction_L2(
    generated_data_all_array, save_result: Callable[[str, np.ndarray], bool]
):
    # data preparation
    (volumes, poses, psf, groundtruth), compute_device = generated_data_all_array
    device = get_device(volumes)
    xp = array_namespace(volumes)
    volumes = xp.stack((volumes,) * 3, axis=1)
    psf = xp.stack((psf,) * 3, axis=0)
    dtype = volumes.dtype
    noise = xp.asarray(
        np.random.randn(*volumes.shape) * 0.2, device=device, dtype=dtype
    )
    volumes = volumes + noise
    lambda_ = xp.asarray(1.0)

    # reconstruction of 3-channel input should be the same as 3 mono-channel
    # input reconstrucions
    recon_multichannel = reconstruction_L2(
        volumes, psf, poses, lambda_, device=compute_device, multichannel=True
    )
    recon = xp.stack(
        [
            reconstruction_L2(
                volumes[:, i, ...],
                psf[i, ...],
                poses,
                lambda_,
                device=compute_device,
                multichannel=False,
            )
            for i in range(3)
        ]
    )

    save_result("reconstruction_multichannel", recon_multichannel)
    save_result("reconstruction", recon)

    assert recon_multichannel.shape == recon.shape
    assert_allclose(recon_multichannel, recon, atol=1e-4)


@pytest.mark.parametrize(
    "generated_data_all_array", testing_libs, indirect=True, ids=ids
)
def test_symmetry_reconstruction_L2(
    generated_data_all_array, save_result: Callable[[str, np.ndarray], bool]
):
    k = 9
    (volumes, poses, psf, groundtruth), device = generated_data_all_array
    xp = array_namespace(volumes)
    lambda_ = xp.asarray(1.0)
    poses_sym = xp.stack((poses,) * k)  # useless symmetry
    recon_sym = reconstruction_L2(
        volumes, psf, poses_sym, lambda_, symmetry=True, device=device
    )
    recon = reconstruction_L2(
        xp.concat((volumes,) * k, axis=0),
        psf,
        xp.concat((poses,) * k, axis=0),
        lambda_,
        symmetry=False,
        device=device,
    )

    save_result("reconstruction_sym", recon_sym)
    save_result("reconstruction", recon)

    assert recon_sym.shape == volumes.shape[-3:]
    assert_allclose(recon_sym, recon, atol=1e-4)


@pytest.mark.parametrize(
    "generated_data_all_array", testing_libs, indirect=True, ids=ids
)
def test_symmetry_reconstruction_L2_2(
    generated_data_all_array: tuple["Array", ...],
    save_result: Callable[[str, np.ndarray], bool],
):
    """reconstruction_L2 of 1 particle with angles that have been symmetrized
    should get approximately the same result as a simple rotation
    """
    (volumes, groundtruth_poses, psf, groundtruth), device = generated_data_all_array
    xp = array_namespace(volumes)
    lambda_ = xp.asarray(1.0)

    # select particle 0
    volume = volumes[0, ...][None, ...]
    pose = groundtruth_poses[0, ...][None, ...]

    # create symmetrical poses
    pose_sym = xp.stack((pose,) * 9)
    pose_sym[:, 0, :3] = symmetrize_angles(pose[0, :3], symmetry=9, degrees=True)

    # reconstruct
    recon_sym = reconstruction_L2(
        volume, psf, pose_sym, lambda_, symmetry=True, device=device
    )

    # compare with simple rotation
    rot = affine_transform(
        volume[0, ...],
        xp.asarray(
            get_transform_matrix(
                volume.shape[1:], pose[0, :3], pose[0, 3:], degrees=True
            ),
            dtype=volume.dtype,
        ),
        order=1,
    )

    # save and assert
    save_result("reconstruction_sym", recon_sym)
    save_result("simple_rot", rot)
    assert_volumes_aligned(recon_sym, rot)


@pytest.mark.parametrize(
    "generated_data_all_array", testing_libs, indirect=True, ids=ids
)
def test_reconstruction_L2_simple(
    generated_data_all_array: tuple["Array", ...],
    save_result: Callable[[str, np.ndarray], bool],
):
    """Do a reconstruction and compare if it's aligned with the groundtruth"""
    (volumes, groundtruth_poses, psf, groundtruth), device = generated_data_all_array
    xp = array_namespace(volumes)
    lambda_ = xp.asarray(1.0)
    reconstruction = reconstruction_L2(
        volumes, psf, groundtruth_poses, lambda_, device=device
    )

    save_result("reconstruction", reconstruction)
    save_result("groundtruth", groundtruth)

    assert_volumes_aligned(reconstruction, groundtruth, atol=1)


@pytest.mark.parametrize(
    "generated_data_all_array", testing_libs, indirect=True, ids=ids
)
def test_reconstruction_L2_symmetry(
    generated_data_all_array: tuple["Array", ...],
    save_result: Callable[[str, np.ndarray], bool],
):
    """Do a reconstruction with symmetry and compare if it's aligned with groundtruth"""
    (volumes, groundtruth_poses, psf, groundtruth), device = generated_data_all_array
    xp = array_namespace(volumes)
    lambda_ = xp.asarray(1.0)
    euler_angles_sym = symmetrize_angles(
        groundtruth_poses[:, :3], symmetry=9, degrees=True
    )
    gt_poses_sym = xp.concat(
        (euler_angles_sym, xp.zeros_like(euler_angles_sym)), axis=2
    )
    gt_poses_sym[:, :, 3:] = groundtruth_poses[:, None, 3:]  # shape (N, k, 6)
    gt_poses_sym = xp.permute_dims(gt_poses_sym, (1, 0, 2))  # shape (k, N, 6)

    reconstruction = reconstruction_L2(
        volumes, psf, gt_poses_sym, lambda_, symmetry=True, device=device
    )

    save_result("reconstruction", reconstruction)
    save_result("groundtruth", groundtruth)

    assert_volumes_aligned(reconstruction, groundtruth, atol=1)


@pytest.mark.parametrize(
    "generated_data_all_array", testing_libs, indirect=True, ids=ids
)
def test_reconstruction_L2_symmetry_1vol_iso(
    generated_data_all_array: tuple["Array", ...],
    save_result: Callable[[str, np.ndarray], bool],
):
    """Do a reconstruction with 1 volume in 2 ways:
        - 1 volume, 9 poses, reconstruction_L2 with symmetry=True
        - 9x 1 volume, 9 poses, reconstruction_L2 with symmetry=False

    The results must be the same.
    """
    (volumes, groundtruth_poses, psf, groundtruth), device = generated_data_all_array
    xp = array_namespace(volumes)
    lambda_ = xp.asarray(1.0)
    euler_angles_sym = symmetrize_angles(
        groundtruth_poses[:, :3], symmetry=9, degrees=True
    )
    gt_poses_sym = xp.concat(
        (euler_angles_sym, xp.zeros_like(euler_angles_sym)), axis=2
    )
    gt_poses_sym[:, :, 3:] = groundtruth_poses[:, None, 3:]  # shape (N, k, 6)
    gt_poses_sym = xp.permute_dims(gt_poses_sym, (1, 0, 2))  # shape (k, N, 6)

    reconstruction_sym = reconstruction_L2(
        volumes[:1, ...],
        psf,
        gt_poses_sym[:, :1, :],
        lambda_,
        symmetry=True,
        device=device,
    )
    volume0_repeated = xp.stack((volumes[0, ...],) * 9)
    reconstruction = reconstruction_L2(
        volume0_repeated,
        psf,
        gt_poses_sym[:, 0, :],
        lambda_,
        symmetry=False,
        device=device,
    )

    save_result("reconstruction_sym=True", reconstruction_sym)
    save_result("reconstruction_sym=False", reconstruction)

    assert_allclose(reconstruction_sym, reconstruction, rtol=0.01, atol=1e-5)


@pytest.mark.parametrize(
    "generated_data_all_array", testing_libs, indirect=True, ids=ids
)
def test_reconstruction_L2_symmetry_Nvol_iso(
    generated_data_all_array: tuple["Array", ...],
    save_result: Callable[[str, np.ndarray], bool],
):
    """Do a reconstruction with N volumes in 2 ways:
        - N volumes, Nx9 poses, reconstruction_L2 with symmetry=True
        - 9xN volumes, Nx9 poses, reconstruction_L2 with symmetry=False

    The results must be the same.
    """
    (volumes, groundtruth_poses, psf, groundtruth), device = generated_data_all_array
    xp = array_namespace(volumes)
    lambda_ = xp.asarray(1.0)
    euler_angles_sym = symmetrize_angles(
        groundtruth_poses[:, :3], symmetry=9, degrees=True
    )
    gt_poses_sym = xp.concat(
        (euler_angles_sym, xp.zeros_like(euler_angles_sym)), axis=2
    )
    gt_poses_sym[:, :, 3:] = groundtruth_poses[:, None, 3:]  # shape (N, k, 6)
    gt_poses_sym = xp.permute_dims(gt_poses_sym, (1, 0, 2))  # shape (k, N, 6)

    reconstruction_sym = reconstruction_L2(
        volumes, psf, gt_poses_sym, lambda_, symmetry=True, device=device
    )
    volumes_repeated = xp.concat(
        [xp.stack((volumes[i, ...],) * 9) for i in range(volumes.shape[0])]
    )
    reconstruction = reconstruction_L2(
        volumes_repeated,
        psf,
        xp.reshape(xp.permute_dims(gt_poses_sym, (1, 0, 2)), (-1, 6)),
        lambda_,
        symmetry=False,
        device=device,
    )

    save_result("reconstruction_sym=True", reconstruction_sym)
    save_result("reconstruction_sym=False", reconstruction)

    assert_allclose(reconstruction_sym, reconstruction, rtol=0.01, atol=1e-6)


#############################
# Test convolution_matching #
#############################

gpu_libs, gpu_ids = zip(
    *filter(
        lambda x: x[0][1] == "cuda" or "cupy" in x[0][0].__name__,
        zip(testing_libs, ids),
    )
)


# This test requires ~12GB VRAM
@pytest.mark.slow
@pytest.mark.parametrize(
    "generated_data_all_array", gpu_libs, indirect=True, ids=gpu_ids
)
def test_memory_convolution_matching_poses_grid(generated_data_all_array):
    """Test if a out of memory error occurs"""
    (volumes, groundtruth_poses, psf, groundtruth), device = generated_data_all_array
    xp = array_namespace(volumes)
    # 10 volumes of size 50^3, float32 -> 5 Mo
    # M = 10_000 -> 50 Go to transfer to GPU
    # We split the computation in batch_size
    for M in [1, 10, 100, 1_000, 10_000]:
        potential_poses = xp.asarray(np.random.randn(M, 6), dtype=volumes.dtype)

        best_poses, errors = convolution_matching_poses_grid(
            groundtruth[None],
            volumes[:, None],
            psf[None],
            potential_poses,
            device=device,
            batch_size=256,
        )

        assert best_poses.shape == (volumes.shape[0], 6)
        assert errors.shape == (volumes.shape[0],)


@pytest.mark.parametrize("xp,device", gpu_libs, ids=gpu_ids)
def test_shapes_convolution_matching_poses_grid(xp, device):
    M, d = 5, 6
    N, C, D, H, W = 10, 2, 32, 32, 32
    reference = xp.asarray(np.random.randn(C, D, H, W), device=device)
    volumes = xp.asarray(np.random.randn(N, C, D, H, W), device=device)
    psf = xp.asarray(np.random.randn(C, D, H, W), device=device)
    potential_poses = xp.asarray(np.random.randn(M, d), device=device)

    best_poses, errors = convolution_matching_poses_grid(
        reference, volumes, psf, potential_poses
    )

    assert best_poses.shape == (N, d)
    assert errors.shape == (N,)


# TODO faire les tests matlab
# def test_matlab_convolution_matching_poses_refined():
#     def as_tensor(x):
#         return torch.as_tensor(x, dtype=torch.float64, device="cuda")
#
#     # Load Matlab data
#     data_path = \
#       os.path.join(os.path.dirname(__file__), "data", "convolution_matching")
#     potential_poses_ = loadmat(os.path.join(data_path, "bigListPoses.mat"))[
#         "bigListPoses"
#     ]
#     volumes = np.stack(
#         loadmat(os.path.join(data_path, "inVols.mat"))["inVols"][:, 0]
#     ).transpose(0, 3, 2, 1)
#     best_poses_matlab = loadmat(os.path.join(data_path, "posesNew.mat"))["posesNew"][
#         :, [0, 1, 2, 5, 3, 4]
#     ]
#     best_poses_matlab[:, 3:] *= -1
#     psf = loadmat(os.path.join(data_path, "psf.mat"))["psf"].transpose(2, 1, 0)
#     reference = loadmat(os.path.join(data_path, "recon.mat"))["recon1"].transpose(
#         2, 1, 0
#     )
#
#     potential_poses_, volumes, best_poses_matlab, psf, reference = map(
#         as_tensor, [potential_poses_, volumes, best_poses_matlab, psf, reference]
#     )
#
#     N, M, _ = potential_poses_.shape
#     potential_poses = as_tensor(torch.zeros((N, M, 6)))
#     potential_poses[:, :, :3] = potential_poses_
#
#     best_poses, _ = convolution_matching_poses_refined(
#         reference, volumes, psf, potential_poses
#     )
#
#     eps = 1e-2
#     assert ((best_poses - best_poses_matlab) < eps).all()


@pytest.mark.parametrize("xp,device", gpu_libs, ids=gpu_ids)
def test_shapes_convolution_matching_poses_refined(xp, device):
    M, d = 5, 6
    N, C, D, H, W = 10, 2, 32, 32, 32
    reference = xp.asarray(np.random.randn(C, D, H, W), device=device)
    volumes = xp.asarray(np.random.randn(N, C, D, H, W), device=device)
    psf = xp.asarray(np.random.randn(C, D, H, W), device=device)
    potential_poses = xp.asarray(np.random.randn(N, M, d), device=device)

    best_poses, errors = convolution_matching_poses_refined(
        reference, volumes, psf, potential_poses, device=device, batch_size=2048
    )

    assert best_poses.shape == (N, d)
    assert errors.shape == (N,)


###################
# Test find_angle #
###################


@pytest.mark.slow
@pytest.mark.parametrize("xp,device", gpu_libs, ids=gpu_ids)
def test_shapes_find_angles_grid(xp, device):
    N, C, D, H, W = 15, 2, 32, 32, 32
    reconstruction = xp.asarray(np.random.randn(C, D, H, W), device=device)
    patches = xp.asarray(np.random.randn(N, C, D, H, W), device=device)
    psf = xp.asarray(np.random.randn(C, D, H, W), device=device)

    best_poses, errors = find_angles_grid(
        xp, reconstruction, patches, psf, precision=10
    )

    assert best_poses.shape == (N, 6)
    assert errors.shape == (N,)


###############
# Test refine #
###############


# batch_size is adjusted for ~12GB VRAM
@pytest.mark.slow
@pytest.mark.parametrize("xp,device", gpu_libs, ids=gpu_ids)
def test_refine_shapes(xp, device):
    N, C, D, H, W = 15, 2, 32, 32, 32
    patches = xp.asarray(np.random.randn(N, C, D, H, W))
    psf = xp.asarray(np.random.randn(C, D, H, W))
    guessed_poses = xp.asarray(np.random.randn(N, 6))

    S = 2
    steps = [(S * S, S), S * S * S]
    ranges = [0, 40]
    recon, poses = refine(
        patches, psf, guessed_poses, steps, ranges, device=device, batch_size=256
    )

    assert recon.shape == patches[0].shape
    assert poses.shape == guessed_poses.shape


@pytest.mark.slow
@pytest.mark.parametrize("xp,device", gpu_libs, ids=gpu_ids)
def test_refine_initial_vol_shapes(xp, device):
    N, C, D, H, W = 15, 2, 32, 32, 32
    patches = xp.asarray(np.random.randn(N, C, D, H, W))
    psf = xp.asarray(np.random.randn(C, D, H, W))
    guessed_poses = xp.asarray(np.random.randn(N, 6))
    initial_volume = xp.asarray(np.random.randn(C, D, H, W))

    S = 2
    steps = [(S * S, S), S * S * S]
    ranges = [0, 40]
    recon, poses = refine(
        patches,
        psf,
        guessed_poses,
        steps,
        ranges,
        device=device,
        batch_size=256,
        initial_volume=initial_volume,
    )

    assert recon.shape == patches[0].shape
    assert poses.shape == guessed_poses.shape


@pytest.mark.parametrize(
    "generated_data_all_array", gpu_libs, indirect=True, ids=gpu_ids
)
def test_refine_easy(
    generated_data_all_array: tuple["Array", ...],
    poses_with_noise: "Array",
    save_result: Callable[[str, np.ndarray], bool],
):
    poses = poses_with_noise

    # data preparation
    (volumes, groundtruth_poses, psf, groundtruth), compute_device = (
        generated_data_all_array
    )
    xp = array_namespace(volumes)

    lambda_ = xp.asarray(1.0, device=get_device(volumes))
    poses_sym = xp.permute_dims(symmetrize_poses(poses, 9), (1, 0, 2))
    initial_reconstruction = reconstruction_L2(
        volumes,
        psf,
        poses_sym,
        lambda_,
        symmetry=True,
        device=compute_device,
        batch_size=1,
    )

    S = 5
    A = 5 * 2
    steps = [(A**2, 5)] + [S] * 7  # 7.25° axis precision; 4° sym precision
    ranges = [
        0,
    ] + [10, 5, 5, 2, 2, 1, 1]
    reconstruction, best_poses = refine(
        volumes[:, None],
        psf[None],
        poses,
        steps,
        ranges,
        symmetry=9,
        lambda_=1e-2,
        device=compute_device,
        batch_size=256,
    )
    reconstruction = reconstruction[0]

    rot_dist_deg1, trans_dist_pix1 = distance_family_poses(
        best_poses, groundtruth_poses, symmetry=9
    )
    rot_dist_deg2, trans_dist_pix2 = distance_family_poses(
        poses, groundtruth_poses, symmetry=9
    )

    save_result("initial_reconstruction", initial_reconstruction)
    save_result("final_reconstruction", reconstruction)

    assert median(rot_dist_deg1) < median(rot_dist_deg2) and median(
        trans_dist_pix1
    ) < median(trans_dist_pix2)


@pytest.mark.parametrize(
    "generated_data_all_array", gpu_libs, indirect=True, ids=gpu_ids
)
def test_refine_easy_multichannel(
    generated_data_all_array: tuple["Array", ...],
    poses_with_noise: "Array",
    save_result: Callable[[str, np.ndarray], bool],
):
    poses = poses_with_noise

    # data preparation
    (volumes, groundtruth_poses, psf, groundtruth), compute_device = (
        generated_data_all_array
    )
    device = get_device(volumes)
    xp = array_namespace(volumes)
    volumes = xp.stack((volumes,) * 3, axis=1)
    psf = xp.stack((psf,) * 3, axis=0)
    dtype = volumes.dtype
    noise = xp.asarray(
        np.random.randn(*volumes.shape) * 0.2, device=device, dtype=dtype
    )
    volumes = volumes + noise
    lambda_ = xp.asarray(1.0)

    lambda_ = xp.asarray(1.0, device=get_device(volumes))
    poses_sym = xp.permute_dims(symmetrize_poses(poses, 9), (1, 0, 2))
    initial_reconstruction = reconstruction_L2(
        volumes,
        psf,
        poses_sym,
        lambda_,
        symmetry=True,
        device=compute_device,
        batch_size=1,
        multichannel=True,
    )

    S = 5
    A = 5 * 2
    steps = [(A**2, 5)] + [S] * 7  # 7.25° axis precision; 4° sym precision
    ranges = [
        0,
    ] + [10, 5, 5, 2, 2, 1, 1]
    reconstruction, best_poses = refine(
        volumes,
        psf,
        poses,
        steps,
        ranges,
        symmetry=9,
        lambda_=1e-2,
        device=compute_device,
        batch_size=256,
    )

    rot_dist_deg1, trans_dist_pix1 = distance_family_poses(
        best_poses, groundtruth_poses, symmetry=9
    )
    rot_dist_deg2, trans_dist_pix2 = distance_family_poses(
        poses, groundtruth_poses, symmetry=9
    )

    save_result("initial_reconstruction", initial_reconstruction)
    save_result("final_reconstruction", reconstruction)

    assert median(rot_dist_deg1) < median(rot_dist_deg2) and median(
        trans_dist_pix1
    ) < median(trans_dist_pix2)



from .evaluate import evaluate_picking, evaluate_tilt
from .postprocess import postprocess
from .predict import predict_picking, predict_tilt

__all__ = [predict_picking, predict_tilt, postprocess, evaluate_picking, evaluate_tilt]



import argparse
import os

from ..viewers import show_particles


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("files", type=str, nargs="+", help="paths to the particles")
    parser.add_argument(
        "--spacing", type=float, default=None, nargs="+", help="Voxel size (ZYX)"
    )
    args = parser.parse_args()

    image_paths = list(map(os.path.abspath, args.files))

    show_particles(image_paths, args.spacing)


if __name__ == "__main__":
    main()



import os
import pickle
from time import time

import numpy as np
import torch
from scipy import ndimage as ndii
from scipy.signal import correlate
from utils import inverse_affine_transform, pad_to_size, reconstruction_L2
from utils.loading import load_annotations, load_array


def load_views(views_path):
    _, ext = os.path.splitext(views_path)
    if ext == ".csv":  # DONT USE CSV NOT WORKING
        views_ = load_annotations(views_path)
        views = views_[:, 2].astype(int)
        patches_names = np.array(
            [
                os.path.splitext(im_name)[0]
                + "_"
                + patch_index
                + os.path.splitext(im_name)[1]
                for im_name, patch_index in views_[:, [0, 1]]
            ]
        )
    elif ext == ".pickle":
        with open(views_path, "rb") as f:
            views_ = pickle.load(f)
        views = np.concatenate([views_[image_name][0] for image_name in views_.keys()])
        patches_names = np.concatenate(
            [views_[image_name][2] for image_name in views_.keys()]
        )

    return views, patches_names


def load_patches(crop_dir, patches_names):
    patches = [
        load_array(os.path.join(crop_dir, p)) for p in patches_names
    ]  # (N,z,y,x)
    return np.stack(patches, axis=0)


def normalize_patches(patches):
    N = patches.shape[0]
    patch_shape = patches.shape[1:]
    flatten_patches = patches.reshape(N, -1)
    min_patch, max_patch = flatten_patches.min(axis=1), flatten_patches.max(axis=1)
    min_patch = min_patch.reshape(tuple([N] + [1] * len(patch_shape)))
    max_patch = max_patch.reshape(tuple([N] + [1] * len(patch_shape)))
    patches = (patches - min_patch) / (max_patch - min_patch)

    return patches


def polar(image, angles=None, radii=None):
    """Return polar transformed image and log base."""
    shape = image.shape
    center = shape[0] / 2, shape[1] / 2
    if angles is None:
        angles = shape[0]
    if radii is None:
        radii = shape[1]
    theta = np.empty((angles, radii), dtype="float64")
    theta.T[:] = np.linspace(0, 2 * np.pi, angles, endpoint=False)
    # d = radii
    d = np.hypot(shape[0] - center[0], shape[1] - center[1])
    # log_base = 10.0 ** (np.log10(d) / (radii))
    radius = np.empty_like(theta)
    # radius[:] = (
    #    np.power(log_base, np.arange(radii, dtype='float64')) - 1.0
    # )
    radius[:] = np.linspace(0, d, radii)
    x = radius * np.sin(theta) + center[0]
    y = radius * np.cos(theta) + center[1]
    output = np.empty_like(x)
    ndii.map_coordinates(image, [x, y], output=output)
    return output


def get_rotation_angle(im1, im2):
    acf1, acf2 = correlate(im1, im1), correlate(im2, im2)
    polar1 = polar(acf1)
    polar2 = polar(acf2)
    repeated_polar1 = np.concatenate([polar1, polar1], axis=0)
    angular_correlation = correlate(repeated_polar1, polar2, mode="valid")
    x = np.linspace(0, 360, angular_correlation.shape[0])
    best_angle = x[angular_correlation.argmax()]
    return best_angle


def get_translation(im1, im2):
    ccf = correlate(im1, im2, mode="full")
    y, x = np.unravel_index(ccf.argmax(), ccf.shape)
    center = (ccf.shape[0] - 1) / 2, (ccf.shape[1] - 1) / 2
    return ccf[y, x], center[0] - y, center[1] - x


def get_transform(im1, im2):
    angle = get_rotation_angle(im1, im2)
    rotated_patch1 = ndii.rotate(im2, -angle, reshape=False)
    rotated_patch2 = ndii.rotate(im2, -(angle + 180), reshape=False)
    ccf1, ty1, tx1 = get_translation(im1, rotated_patch1)
    ccf2, ty2, tx2 = get_translation(im1, rotated_patch2)
    if ccf1 > ccf2:
        return angle, ty1, tx1
    else:
        return angle + 180, ty2, tx2


polar_dict = {}


def get_polar_special(im):
    global polar_dict
    k = im.tobytes()
    if k in polar_dict:
        return polar_dict[k]
    else:
        p = polar(correlate(im, im))
        polar_dict[k] = p
        return p


def get_rotation_angle_special(im1, im2):
    polar1, polar2 = get_polar_special(im1), get_polar_special(im2)
    repeated_polar1 = np.concatenate([polar1, polar1], axis=0)
    angular_correlation = correlate(repeated_polar1, polar2, mode="valid")
    x = np.linspace(0, 360, angular_correlation.shape[0])
    best_angle = x[angular_correlation.argmax()]
    return best_angle


def get_transform_special(im1, im2):
    angle = get_rotation_angle_special(im1, im2)
    rotated_patch1 = ndii.rotate(im2, -angle, reshape=False)
    rotated_patch2 = ndii.rotate(im2, -(angle + 180), reshape=False)
    ccf1, ty1, tx1 = get_translation(im1, rotated_patch1)
    ccf2, ty2, tx2 = get_translation(im1, rotated_patch2)
    if ccf1 > ccf2:
        return angle, ty1, tx1
    else:
        return angle + 180, ty2, tx2


def transform_patch(patch, R, ty, tx, pad=False):
    if pad:
        h, w = patch.shape
        padw = int(h * (2**0.5 - 1) + 2)
        patch = np.pad(patch, [(padw, padw), (padw, padw)])
    r_patch = ndii.rotate(patch, -R, reshape=False)
    s_patch = ndii.shift(r_patch, np.array([-ty, -tx]))
    if pad:
        s_patch = s_patch[padw:-padw, padw:-padw]
    return s_patch


def build_particle(patches, tree, R, ty, tx):
    r_, ty_, tx_ = tree[3]
    if isinstance(tree[2], list):
        m = tree[0]
        p = patches[tree[1]]
        return (
            transform_patch(
                p + (m - 1) * build_particle(patches, tree[2], r_, ty_, tx_), R, ty, tx
            )
            / m
        )
    else:
        assert tree[0] == 2
        r_, ty_, tx_ = tree[3]
        p = (patches[tree[1]] + transform_patch(patches[tree[2]], r_, ty_, tx_)) / 2
        return transform_patch(p, R, ty, tx)


def fill_params(params, tree, R, ty, tx):
    r_, ty_, tx_ = tree[3]
    params[tree[1]] = (R % 360, ty, tx)
    R_rad = R * np.pi / 180
    if isinstance(tree[2], list):
        fill_params(
            params,
            tree[2],
            R + r_,
            tx * np.sin(R_rad) + ty * np.cos(R_rad) - ty_,
            tx * np.cos(R_rad) - ty * np.sin(R_rad) - tx_,
        )
        # fill_params(params, tree[2], R+r_, ty+ty_, tx+tx_)
    else:
        params[tree[2]] = (
            (R + r_) % 360,
            tx * np.sin(R_rad) + ty * np.cos(R_rad) - ty_,
            tx * np.cos(R_rad) - ty * np.sin(R_rad) - tx_,
        )
        # params[tree[2]] = ((R+r_)%360,  ty+ty_, tx+tx_)


def iterative_alignment(particles, epsilon=1e-2, max_iter=20):
    N, h, w = particles.shape

    pad = int(h * (2**0.5 - 1) + 2)
    # pad = 0
    patches_padded = np.pad(particles, [(0, 0), (pad, pad), (pad, pad)])
    _, H, W = patches_padded.shape

    # Phase A : build an average particle

    patches_param = np.zeros((N, 3))
    order = np.random.choice(np.arange(N), N, replace=False)
    # Center patches
    patches_padded_centered = np.zeros_like(patches_padded)
    for i in range(N):
        y_centermass, x_centermass = ndii.center_of_mass(patches_padded[order[i]])
        center = (H - 1) / 2, (W - 1) / 2
        ty, tx = center[0] - y_centermass, center[1] - x_centermass
        patches_param[i] = [0, -ty, -tx]
        patches_padded_centered[order[i]] = transform_patch(
            patches_padded[order[i]], 0, -ty, -tx
        )

    # Initialize the tree
    tree = [
        2,
        order[0],
        order[1],
        get_transform(patches_padded_centered[order[0]], patches_padded[order[1]]),
    ]

    # Build the tree
    while tree[0] < N:
        m = int(tree[0])
        p_right = build_particle(patches_padded_centered, tree, 0, 0, 0)

        # find the transformation
        p_left = patches_padded_centered[order[m]]
        angle, ty, tx = get_transform(p_left, p_right)

        ## center the particle
        # y_centermass, x_centermass = ndii.center_of_mass(p_right)
        # ty_corr, tx_corr = center[0] - y_centermass, center[1] - x_centermass

        tree = [m + 1, order[m], tree, (angle, ty, tx)]

    # Build the average particle
    avrg_particle = build_particle(patches_padded_centered, tree, 0, 0, 0)
    # plt.imshow(avrg_particle, cmap='gray')
    # plt.show()

    # Get the particles' params from the tree
    patches_param_centered = np.zeros_like(patches_param)
    fill_params(patches_param_centered, tree, 0, 0, 0)
    print(tree)
    # patches_param = patches_param + patches_param_centered

    # Phase B : stabilize the average particle
    order = np.arange(N)
    distance = np.array([3, 3, 3])
    j = 0
    while distance.sum() > epsilon and j < max_iter:
        j += 1
        np.random.shuffle(order)  # shuffle the particles
        new_patches_param = np.array([(0, 0, 0) for i in range(N)])
        last_transforms = [None for i in range(N)]
        for i in range(N):
            # p is the patch to remove
            p = patches_padded_centered[order[i]]
            transformed_p = transform_patch(
                p, *patches_param_centered[order[i]]
            )  # if last_transforms[order[i]] is None else last_transforms[order[i]]
            print(order[i], patches_param_centered[order[i]])

            # plt.figure(figsize=(10,5))
            # plt.subplot(151)
            # plt.imshow(avrg_particle, cmap='gray')
            # remove the patch p from the average
            avrg_particle = (N * avrg_particle - transformed_p) / (N - 1)
            # plt.subplot(152)
            # plt.imshow(p, cmap='gray')
            # plt.subplot(153)
            # plt.imshow(transformed_p, cmap='gray')
            # plt.subplot(154)
            # plt.imshow(avrg_particle, cmap='gray')

            # align the patch p and the new average
            R, ty, tx = get_transform(avrg_particle, p)

            # save the new params of patch p
            new_patches_param[order[i]] = np.array([R, ty, tx])

            # reconstruct the average particle
            new_transformed_p = transform_patch(p, *new_patches_param[order[i]])
            # plt.subplot(155)
            # plt.imshow(new_transformed_p, cmap='gray')
            # plt.show()
            last_transforms[order[i]] = new_transformed_p
            avrg_particle = (new_transformed_p + (N - 1) * avrg_particle) / N

        distances = np.square(
            (new_patches_param - patches_param) / np.array([360, 1, 1])
        )
        distance = distances.mean(axis=0) ** 0.5
        patches_param_centered = new_patches_param

    unpadded_avrg_particle = avrg_particle[pad:-pad, pad:-pad]
    patches_param[:, 0] = patches_param[:, 0] % 360

    variance_map = sum(
        [
            (transform_patch(particles[i], *patches_param[i]) - unpadded_avrg_particle)
            ** 2
            for i in range(N)
        ]
    ) / (N - 1)

    return patches_param, unpadded_avrg_particle, variance_map


def iterative_suppression(patches, n):
    N = patches.shape[0]
    assert n <= N
    n_iter = int(np.ceil(np.log2(N / n)))
    mask_keep = np.ones((N), dtype=bool)
    good_patches = patches.copy()
    for i in range(n_iter):
        params, avrg_particle, variance_map = iterative_alignment(good_patches)
        std = np.array(
            [
                (
                    (transform_patch(good_patches[i], *params[i]) - avrg_particle) ** 2
                ).mean()
                ** 0.5
                for i in range(good_patches.shape[0])
            ]
        )
        if i < n_iter - 1:  # not last iteration
            mask_keep[mask_keep] = std <= np.median(
                std
            )  # we keep half of the particles
        else:
            x = np.where(mask_keep)[0]
            mask_keep[x[np.argsort(std)[n:]]] = False

        good_patches = patches[mask_keep]

    params, avrg_particle, variance_map = iterative_alignment(good_patches)

    return mask_keep, params, avrg_particle, variance_map


def get_first_euler_angle_side(side_particles, top_particles, psf, lambda_=1e-2):
    device = side_particles.device
    dtype = side_particles.dtype

    def as_tensor(x):
        return torch.as_tensor(x, device=device, dtype=dtype)

    t0 = time()
    side_param, _, _ = iterative_alignment(side_particles.sum(dim=1).cpu().numpy())
    t1 = time()
    top_param, _, _ = iterative_alignment(top_particles.sum(dim=1).cpu().numpy())
    t2 = time()

    side_poses = torch.zeros((side_particles.size(0), 6), device=device)
    top_poses = torch.zeros((top_particles.size(0), 6), device=device)

    # Known a priori
    side_poses[:, 1] = 90
    top_poses[:, 1] = 0
    top_poses[:, 0] = 0

    # Found with iterative alignment
    # relative 3rd euler angle of top particles
    # top_poses[:, 2] = as_tensor(top_param[:, 0])
    # relative 1st euler angle of side particles
    side_poses[:, 0] = (-as_tensor(side_param[:, 0])) % 360
    # translations in the XY plane
    side_poses[:, [4, 5]] = as_tensor(side_param[:, [1, 2]])
    top_poses[:, [4, 5]] = as_tensor(top_param[:, [1, 2]])

    # Z-axis translation with center of mass
    side_centers_of_mass = (
        np.array(
            list(map(lambda x: ndii.center_of_mass(x)[0], side_particles.cpu().numpy()))
        )
        - side_particles.size(1) / 2
    )
    side_poses[:, 3] = as_tensor(side_centers_of_mass)
    top_centers_of_mass = (
        np.array(
            list(map(lambda x: ndii.center_of_mass(x)[0], top_particles.cpu().numpy()))
        )
        - side_particles.size(1) / 2
    )
    top_poses[:, 3] = as_tensor(top_centers_of_mass)

    # Find first euler angle of side particles
    step = 1
    delta = torch.arange(0, 360, step)[:, None]
    M = delta.size(0)
    side_poses_expanded = side_poses[None].repeat(M, 1, 1)
    side_poses_expanded[:, :, 0] += as_tensor(delta)
    top_poses_expanded = top_poses[None].repeat(M, 1, 1)
    poses_expanded = torch.cat((side_poses_expanded, top_poses_expanded), dim=1)
    particles = torch.cat((side_particles, top_particles), dim=0)

    lambda_ = as_tensor(lambda_)
    t3 = time()
    recon, den = reconstruction_L2(
        particles, psf, poses_expanded, lambda_[None].repeat(M)
    )
    # dt = [0 for _ in range(7)]
    t4 = time()

    MSE = torch.empty((recon.size(0),))
    recon_fft = torch.fft.fftn(torch.fft.fftshift(recon, dim=(1, 2, 3)), dim=(1, 2, 3))
    # recon_fft = torch.fft.fftn(recon, dim=(1,2,3))
    for i in range(recon.size(0)):
        patches_transformed = inverse_affine_transform(
            particles[:, None], poses_expanded[i]
        )[:, 0]
        patches_transformed_fft = torch.fft.fftn(
            torch.fft.fftshift(patches_transformed, dim=(1, 2, 3)), dim=(1, 2, 3)
        )
        # patches_transformed_fft = torch.fft.fftn(
        #    patches_transformed,
        #    dim=(1,2,3)
        # )
        poses_no_tvec = poses_expanded[i]
        poses_no_tvec[:, 3:] = 0
        psf_transformed = inverse_affine_transform(
            psf.repeat(poses_expanded.size(1), 1, 1, 1, 1), poses_no_tvec
        )[:, 0]

        psf_transformed_fft = torch.fft.fftn(
            torch.fft.fftshift(
                pad_to_size(
                    psf_transformed, psf_transformed.shape[:1] + particles.shape[-3:]
                ),
                dim=(1, 2, 3),
            ),
            dim=(1, 2, 3),
        )
        # psf_transformed_fft = torch.fft.fftn(
        #    pad_to_size(
        #       psf_transformed, psf_transformed.shape[:1]+particles.shape[-3:]
        #    ),
        #    dim=(1,2,3)
        # )
        recons_conv_fft = recon_fft[i] * psf_transformed_fft

        MSE[i] = (
            recons_conv_fft - patches_transformed_fft
        ).square().sum() / particles.size(0)

    best_i = MSE.argmin()
    best_recon = recon[best_i]
    side_poses[:, 0] = side_poses_expanded[best_i, :, 0]
    t5 = time()

    print(f"1st iterative alignment : {t1-t0:.2f}s")
    print(f"2nd iterative alignment : {t2-t1:.2f}s")
    print(f"create poses : {t3-t2:.2f}s")
    print(f"L2 Reconstructions : {t4-t3:.2f}s")
    print(f"MSEs Computations : {t5-t4:.2f}s")

    return side_poses, top_poses, best_recon, MSE[best_i]



import napari
import numpy as np
from napari_bbox import BoundingBoxLayer

from spfluo.visualisation.multiple_viewer_widget import add_orthoviewer_widget


def create_multiviewer_with_data(viewer):
    _, widget, _ = add_orthoviewer_widget(viewer)
    viewer.window.add_dock_widget(widget)
    image_layer = viewer.add_image(np.random.randn(10, 100, 100))

    return widget, image_layer


def add_bbox_layer(viewer):
    bbox_layer = BoundingBoxLayer(ndim=3)
    viewer.add_layer(bbox_layer)
    bbox_layer.add(
        [
            [1, 1, 1],
            [10, 1, 1],
            [1, 10, 1],
            [1, 1, 10],
            [10, 10, 1],
            [10, 1, 10],
            [1, 10, 10],
            [10, 10, 10],
        ]
    )
    return bbox_layer


def test_data(make_napari_viewer_proxy):
    viewer: napari.Viewer = make_napari_viewer_proxy(show=True)
    widget, image_layer = create_multiviewer_with_data(viewer)
    add_bbox_layer(viewer)

    assert (
        len(widget.viewer.layers)
        == len(widget.viewer_model1.layers)
        == len(widget.viewer_model2.layers)
    )
    np.testing.assert_allclose(
        widget.viewer.layers[1].data, widget.viewer_model1.layers[1].data
    )
    np.testing.assert_allclose(
        widget.viewer.layers[1].data, widget.viewer_model2.layers[1].data
    )


def test_data_after_move(make_napari_viewer_proxy):
    viewer: napari.Viewer = make_napari_viewer_proxy(show=True)
    widget, image_layer = create_multiviewer_with_data(viewer)
    bbox_layer = add_bbox_layer(viewer)

    # move bbox in main viewer
    bbox_layer.data = [bbox_layer.data[0] + np.asarray([0.0, 5.0, 5.0])]

    assert (
        len(widget.viewer.layers)
        == len(widget.viewer_model1.layers)
        == len(widget.viewer_model2.layers)
    )
    np.testing.assert_allclose(
        widget.viewer.layers[1].data, widget.viewer_model1.layers[1].data
    )
    np.testing.assert_allclose(
        widget.viewer.layers[1].data, widget.viewer_model2.layers[1].data
    )


def test_remove(make_napari_viewer_proxy):
    viewer: napari.Viewer = make_napari_viewer_proxy(show=False)
    widget, image_layer = create_multiviewer_with_data(viewer)
    bbox_layer = add_bbox_layer(viewer)

    # select+remove
    bbox_layer.selected_data = {0}
    bbox_layer.remove_selected()

    assert (
        len(widget.viewer.layers)
        == len(widget.viewer_model1.layers)
        == len(widget.viewer_model2.layers)
    )
    np.testing.assert_allclose(
        widget.viewer.layers[1].data, widget.viewer_model1.layers[1].data
    )
    np.testing.assert_allclose(
        widget.viewer.layers[1].data, widget.viewer_model2.layers[1].data
    )


def test_remove2(make_napari_viewer_proxy):
    viewer: napari.Viewer = make_napari_viewer_proxy(show=True)
    widget, image_layer = create_multiviewer_with_data(viewer)
    bbox_layer = add_bbox_layer(viewer)
    bbox_layer.add(
        [
            [12, 12, 12],
            [20, 12, 12],
            [12, 20, 12],
            [12, 12, 20],
            [20, 20, 12],
            [20, 12, 20],
            [12, 20, 20],
            [20, 20, 20],
        ]
    )

    # select+remove
    bbox_layer.selected_data = {0}
    bbox_layer.remove_selected()

    assert (
        len(widget.viewer.layers)
        == len(widget.viewer_model1.layers)
        == len(widget.viewer_model2.layers)
    )
    np.testing.assert_allclose(
        widget.viewer.layers[1].data, widget.viewer_model1.layers[1].data
    )
    np.testing.assert_allclose(
        widget.viewer.layers[1].data, widget.viewer_model2.layers[1].data
    )



"""Basically nothing more than classic Pytorch dataloaders creation. This only external call
should be to the 'get_dataloaders' function. Other functions defined here should only be used
here internally by the 'get_dataloaders' function.

The data folder must has the following structure:

rootdir
|
|__train
|    |___ cropped
|            |___ positive
|            |___ negative
|__ val
     |___ cropped
             |___ positive
             |___ negative
"""

import os
import pickle
from typing import List, Tuple

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset

from ..utils import load_array
from .augment import get_augment_policy

# +------------------------------------------------------------------------------------------+ #
# |                                          DATASET                                         | #
# +------------------------------------------------------------------------------------------+ #


def get_patch_paths(
    rootdir: str, num_pos_samples: int = None, shuffle: bool = False
) -> List[str]:
    all_paths = []
    for subdir in ["positive", "negative"]:
        datadir = os.path.join(rootdir, subdir)
        paths = sorted(
            list(map(lambda x: os.path.join(datadir, x), os.listdir(datadir)))
        )
        if shuffle:
            np.random.shuffle(paths)
        all_paths.extend(paths[:num_pos_samples])
    if shuffle:
        np.random.shuffle(all_paths)
    return all_paths


class Trainset(Dataset):
    def __init__(
        self,
        paths: List[str],
        dim: int,
        augment: float,
        size: int = None,
        training: bool = True,
    ) -> None:
        super().__init__()
        self.paths = paths
        self.dim = dim
        patch_size = load_array(self.paths[0]).shape
        self.augment = get_augment_policy(patch_size, p=augment, dim=dim)
        self.training = training
        self.size = size

    @staticmethod
    def get_target(path: str) -> int:
        target = 1 if "positive" in path else 0
        return torch.as_tensor(target).unsqueeze(dim=-1).float()

    def get_image(self, path: str):
        image = load_array(path)
        if self.dim == 2:
            image = image.sum(axis=0)
            image = (image - image.min()) / (image.max() - image.min())
        if self.training:
            image = self.augment(image=image)["image"]
        image = (image - image.min()) / (image.max() - image.min())
        image = np.expand_dims(image, axis=0)  # add channel dim at the beginning
        return torch.from_numpy(image.astype(np.float32))

    def __getitem__(self, index: int) -> Tuple[torch.Tensor]:
        if index >= len(self.paths):
            np.random.shuffle(self.paths)
        index = index % len(self.paths)
        return self.get_image(self.paths[index]), self.get_target(self.paths[index])

    def __len__(self) -> int:
        return self.size if self.size is not None else len(self.paths)


# +------------------------------------------------------------------------------------------+ #
# |                                       EXTERNAL CALL                                      | #
# +------------------------------------------------------------------------------------------+ #


def make_dataloaders(
    rootdir: str,
    output_dir: str,
    batch_size: int,
    num_workers: int,
    num_pos_samples: int,
    shuffle: bool,
    augment: float,
    epoch_size: int,
    dim: int,
) -> Tuple[DataLoader]:
    # train data
    datadir = os.path.join(rootdir, "train", "cropped")
    train_paths = get_patch_paths(datadir, num_pos_samples, shuffle)
    with open(os.path.join(output_dir, "selected_particles.pickle"), "wb") as file:
        pickle.dump(train_paths, file)
    size = batch_size * epoch_size if epoch_size is not None else None
    train_set = Trainset(train_paths, dim, augment, size, training=True)
    # val data
    datadir = os.path.join(rootdir, "val", "cropped")
    val_paths = get_patch_paths(datadir)  # num_pos_samples=None, shuffle=False
    val_set = Trainset(val_paths, dim, augment, training=False)
    loader_kwargs = {"batch_size": batch_size, "num_workers": num_workers}
    train_loader = DataLoader(train_set, **loader_kwargs, shuffle=True)
    val_loader = DataLoader(val_set, **loader_kwargs, shuffle=False)
    return train_loader, val_loader



from typing import Callable, Tuple

import albumentations as A

from . import volumentations as V

# +------------------------------------------------------------------------------------------+ #
# |                                      AUGMENTATION API                                    | #
# +------------------------------------------------------------------------------------------+ #


def augment_3d(patch_size: Tuple[int], p: float) -> V.Compose:
    return V.Compose(
        [
            V.RandomScale((0.8, 1.2), always_apply=True),
            V.PadIfNeeded(patch_size, always_apply=True),
            V.Resize(patch_size, always_apply=True),
            V.RandomRotate90((1, 2)),
            V.Flip(0),
        ],
        p=p,
    )


def augment_2d(patch_size: Tuple[int], p: float) -> Callable:
    size = max(patch_size)  # albumentations doesn't handle non square image.
    return A.Compose(
        [
            A.RandomRotate90(),
            A.Flip(),
            A.ShiftScaleRotate(
                shift_limit=0.2, scale_limit=0.2, rotate_limit=45, p=0.5
            ),
            A.RandomSizedCrop((size - size // 5, size), size, size, p=0.8),
            # A.CoarseDropout(max_holes=5, max_height=8, max_width=8, p=0.5)
        ],
        p=p,
    )


def get_augment_policy(
    patch_size: Tuple[int], p: float = 0.8, dim: int = 2
) -> Callable:
    return augment_2d(patch_size, p) if dim == 2 else augment_3d(patch_size, p)



import argparse

import tifffile

from spfluo.ab_initio_reconstruction.common_image_processing_methods.others import (
    normalize,
)
from spfluo.refinement import refine
from spfluo.utils.array import array_namespace, get_device, to_numpy
from spfluo.utils.loading import read_poses, save_poses
from spfluo.utils.log import base_parser, set_logging_level
from spfluo.utils.read_save_files import (
    read_image,
    read_images_in_folder,
)


def create_parser():
    parser = argparse.ArgumentParser(
        "Refinement",
        parents=[base_parser],
    )

    # Input files
    parser.add_argument("--particles_dir", type=str, required=True)
    parser.add_argument("--psf_path", type=str, required=True)
    parser.add_argument("--guessed_poses_path", type=str, required=True)
    parser.add_argument("--initial_volume_path", type=str, required=False, default=None)
    parser.add_argument("--channel", type=int, required=False, default=None)

    # Output files
    parser.add_argument(
        "--output_reconstruction_path",
        type=str,
        required=False,
        default="./reconstruction.tiff",
    )
    parser.add_argument(
        "--output_poses_path", type=str, required=False, default="./poses.csv"
    )

    # Parameters
    def tuple_of_int(string):
        if "(" in string:
            string = string[1:-1]
        t = tuple(map(int, string.split(",")))
        if len(t) == 2:
            return t
        elif len(t) == 1:
            return t[0]
        else:
            raise TypeError

    parser.add_argument(
        "--steps", nargs="+", action="append", type=tuple_of_int, required=True
    )
    parser.add_argument(
        "--ranges", nargs="+", action="append", type=float, required=True
    )
    parser.add_argument("-l", "--lambda_", type=float, required=False, default=100.0)
    parser.add_argument(
        "--symmetry",
        type=int,
        required=False,
        default=1,
        help="Adds a constraint to the refinement. "
        "The symmetry is cylindrical around the X-axis.",
    )

    parser.add_argument("--dtype", type=str, default="float32")

    # GPU
    parser.add_argument("--gpu", action="store_true")
    parser.add_argument("--minibatch_size", type=int, default=None)

    return parser


def main(args):
    particles, _ = read_images_in_folder(
        args.particles_dir, alphabetic_order=True, gpu=args.gpu, dtype=args.dtype
    )
    xp = array_namespace(particles)
    compute_device = get_device(particles)
    particles = xp.to_device(particles, "cpu")
    host_device = get_device(particles)
    particles = xp.stack([normalize(particles[i]) for i in range(particles.shape[0])])
    if particles.ndim == 4:
        particles = particles[:, None]
    psf = normalize(
        read_image(args.psf_path, xp=xp, device=host_device, dtype=args.dtype)
    )
    if psf.ndim == 3:
        psf = xp.stack((psf,) * particles.shape[1], axis=0)
    if args.initial_volume_path:
        initial_volume = normalize(
            read_image(
                args.initial_volume_path, xp=xp, device=host_device, dtype=args.dtype
            )
        )
        if initial_volume.ndim == 3:
            initial_volume = xp.stack((initial_volume,) * particles.shape[1], axis=0)
    else:
        initial_volume = None
    guessed_poses, names = read_poses(args.guessed_poses_path, alphabetic_order=True)
    guessed_poses = xp.asarray(
        guessed_poses, device=host_device, dtype=getattr(xp, args.dtype)
    )

    if args.channel is not None:
        particles = particles[:, [args.channel]]
        psf = psf[[args.channel]]
        if initial_volume is not None:
            initial_volume = initial_volume[[args.channel]]

    reconstruction, poses = refine(
        particles,
        psf,
        guessed_poses,
        args.steps[0],
        args.ranges[0],
        initial_volume,
        args.lambda_,
        args.symmetry,
        device=compute_device,
        batch_size=args.minibatch_size,
    )

    reconstruction, poses = to_numpy(reconstruction, poses)
    tifffile.imwrite(
        args.output_reconstruction_path,
        reconstruction,
        metadata={"axes": "CZYX"},
    )
    save_poses(args.output_poses_path, poses, names)


if __name__ == "__main__":
    p = create_parser()
    args = p.parse_args()
    set_logging_level(args)
    main(args)



import argparse
import csv
import os
import pickle
import shutil
from glob import glob
from typing import Sequence

import numpy as np

from .modules.annotating import extract_annotations
from .modules.annotating.extract import create_annotation_image
from .modules.posttraining import (
    evaluate_picking,
    evaluate_tilt,
    postprocess,
    predict_picking,
    predict_tilt,
)
from .modules.pretraining import crop, prepare
from .modules.training import train_picking, train_tilt
from .modules.utils import load_annotations, seed_all, send_mail

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # Disable tensorflow INFO and WARNING


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser("Setup pipeline")
    # - General
    all_stages = [
        "extract",
        "prepare",
        "train",
        "predict",
        "postprocess",
        "evaluate",
        "crop",
    ]
    parser.add_argument("--stages", type=str, default=all_stages, nargs="+")
    parser.add_argument(
        "--task", type=str, default="picking", choices=["picking", "tilt"]
    )
    parser.add_argument("--rootdir", type=str, default=None)
    parser.add_argument("--output_dir", type=str, default=None)
    parser.add_argument("--patch_size", type=int, default=None, nargs="+")
    parser.add_argument("--extension", type=str, default="npz")
    parser.add_argument("--dim", type=int, default=2)
    parser.add_argument("--mail", action="store_true")
    # 0 - extract
    parser.add_argument("--slicer3d_dir", type=str, default=None)
    parser.add_argument("--downscale", type=float, default=1.0)
    # I - prepare
    parser.add_argument("--template_pointcloud", type=str, default=None)
    parser.add_argument("--dataset_size", type=int, default=None)
    parser.add_argument("--make_u_masks", action="store_true")
    parser.add_argument("--train_test_split", type=float, default=None)
    parser.add_argument("--train_val_split", type=float, default=None)
    parser.add_argument("--crop_output_dir", type=str, default=None)
    parser.add_argument("--margin", type=int, default=0, nargs="+")
    parser.add_argument("--pos_ratio", type=float, default=1.0)
    parser.add_argument("--positive_only", action="store_true")
    # II - train
    #   1. general setup
    parser.add_argument("--mode", type=str, default="pu", choices=["pu", "fs"])
    parser.add_argument("--epoch_size", type=int, default=None)
    parser.add_argument("--batch_size", type=int, default=64)
    parser.add_argument("--num_workers", type=int, default=16)
    parser.add_argument("--num_epochs", type=int, default=100)
    parser.add_argument("--lr", type=float, default=1e-3)
    #   2. model setup
    parser.add_argument("--network", type=str, default="efficientnet-s")
    parser.add_argument("--width_factor", type=float, default=1.0)
    parser.add_argument("--depth_factor", type=float, default=1.0)
    parser.add_argument("--swa", action="store_true")
    parser.add_argument("--swa_lr", type=float, default=1e-5)
    #   3. data setup
    parser.add_argument("--num_pos_samples", type=int, default=None)
    parser.add_argument("--augment", type=float, default=0.0)
    parser.add_argument("--shuffle", action="store_true")
    parser.add_argument("--selected_particles", type=str, default=None)

    #   4. pu
    parser.add_argument("--radius", type=int, default=None)
    parser.add_argument("--num_particles_per_image", type=int, default=None)
    parser.add_argument("--load_u_masks", action="store_true")
    #   5. autoencoder
    parser.add_argument("--ae", action="store_true")
    parser.add_argument("--eta", type=float, default=0.0)
    parser.add_argument("--learn_eta", action="store_true")
    parser.add_argument("--num_features", type=int, default=(8, 16, 32), nargs=3)
    parser.add_argument("--hidden_dim", type=int, default=128)
    # III - predict
    parser.add_argument("--testdir", type=str, default=None)
    parser.add_argument("--testset_size", type=int, default=None)
    parser.add_argument("--checkpoint", type=str, default=None)
    parser.add_argument("--stride", type=int, default=None, nargs="+")
    parser.add_argument("--predict_on_u_mask", action="store_true")
    # IV - postprocess
    parser.add_argument("--predictions", type=str, default=None)
    parser.add_argument("--nms_threshold", type=float, default=0.1)
    parser.add_argument("--iterative", action="store_true")
    parser.add_argument("--last_step", type=str, default="spatial_average")
    parser.add_argument("--otsu_margin", type=float, default=0.5)
    # V - evaluate
    parser.add_argument("--step_to_evaluate", type=str, default="last_step")
    parser.add_argument("--make_labelled_volumes", action="store_true")
    parser.add_argument("--annotations", type=str, default=None)
    parser.add_argument("--evaluation_threshold", type=float, default=0.2)
    parser.add_argument("--image_name", type=str, default=None)
    # Tilt args
    parser.add_argument("--pretrain", action="store_true")
    parser.add_argument(
        "--crop_dir", type=str, default=None
    )  # cropped particules to take for pretraining
    return parser.parse_args()


def arg_to_tuple_if_needed(arg: Sequence[int]) -> Sequence[int]:
    if isinstance(arg, int):
        arg = 3 * (arg,)
    if isinstance(arg, list) and len(arg) == 1:
        arg = 3 * (arg[0],)
    return arg


def guess_paths(args: argparse.Namespace) -> argparse.Namespace:
    if args.testdir is None and args.rootdir is not None and "predict" in args.stages:
        args.testdir = os.path.join(args.rootdir, "test")
    if args.checkpoint is None and "predict" in args.stages:
        basename = "checkpoint_swa" if args.swa else "checkpoint"
        args.checkpoint = os.path.join(args.output_dir, f"{basename}.pt")
    if args.predictions is None and (
        "postprocess" in args.stages or "evaluate" in args.stages
    ):
        args.predictions = os.path.join(args.output_dir, "predictions.pickle")
    if args.annotations is None and "evaluate" in args.stages:
        args.annotations = os.path.join(args.testdir, "test_coordinates.csv")
    return args


def mail_at_the_end(args: argparse.Namespace) -> None:
    name = args.output_dir.strip("/").split("/")[-1]
    body = f"""
    Job...: {name}
    Task..: {args.task}
    Stages: {', '.join(args.stages)}
    Pipeline finished.
    """
    send_mail("Pipeline", body)


def main(args: argparse.Namespace) -> None:
    # if shuffle is True, we want annotations shuffling to be unique for each run,
    # but the training must be deterministic, so we seed everything but numpy.
    seed_all(seed_numpy=not args.shuffle)
    args = guess_paths(args)
    args.patch_size = arg_to_tuple_if_needed(args.patch_size)
    args.stride = arg_to_tuple_if_needed(args.stride)
    args.margin = arg_to_tuple_if_needed(args.margin)
    if "extract" in args.stages:
        extract_annotations(
            args.slicer3d_dir,
            args.rootdir,
            args.output_dir,
            args.patch_size,
            args.downscale,
        )
    if "prepare" in args.stages:
        prepare(
            args.rootdir,
            args.extension,
            args.template_pointcloud,
            args.dataset_size,
            args.make_u_masks,
            args.train_test_split,
            args.train_val_split,
            args.crop_output_dir,
            args.patch_size,
            args.margin,
            args.pos_ratio,
            args.positive_only,
            args.downscale,
        )
    # + -------------------------------------- PICKING -------------------------------------- + #
    if "train" in args.stages and args.task == "picking":
        train_picking(
            args.mode,
            args.rootdir,
            args.batch_size,
            args.num_workers,
            args.output_dir,
            args.num_epochs,
            args.lr,
            args.network,
            args.width_factor,
            args.depth_factor,
            args.swa,
            args.swa_lr,
            args.patch_size,
            args.num_pos_samples,
            args.shuffle,
            args.augment,
            args.dim,
            args.extension,
            args.epoch_size,
            args.downscale,
            # pu optional params
            args.radius,
            args.num_particles_per_image,
            args.load_u_masks,
            args.ae,
            args.eta,
            args.learn_eta,
            args.num_features,
            args.hidden_dim,
        )
    if "predict" in args.stages and args.task == "picking":
        predict_picking(
            args.network,
            args.checkpoint,
            args.testdir,
            args.predict_on_u_mask,
            args.patch_size,
            args.dim,
            args.stride,
            args.batch_size,
            args.extension,
            args.testset_size,
            args.image_name,
            args.output_dir,
            args.downscale,
        )
    if "postprocess" in args.stages and args.task == "picking":
        postprocess(
            args.predictions,
            args.testdir,
            args.patch_size,
            args.stride,
            args.nms_threshold,
            args.iterative,
            args.last_step,
            args.otsu_margin,
            args.extension,
            args.testset_size,
            args.image_name,
            args.output_dir,
            args.downscale,
        )
    if "evaluate" in args.stages and args.task == "picking":
        evaluate_picking(
            args.predictions,
            args.annotations,
            args.patch_size,
            args.evaluation_threshold,
            args.step_to_evaluate,
            args.make_labelled_volumes,
            args.testdir,
            args.image_name,
            args.output_dir,
        )
    # + --------------------------------------- TILT ---------------------------------------- + #
    if "crop" in args.stages:
        raw_dir = os.path.join(args.rootdir, "raw")

        os.makedirs(raw_dir, exist_ok=True)
        for image_path in glob(os.path.join(args.slicer3d_dir, "*/*.tif")):
            image_name = os.path.basename(image_path)
            shutil.copyfile(image_path, os.path.join(raw_dir, image_name))

        _, ext = os.path.splitext(args.predictions)
        if ext == ".pickle":
            with open(args.predictions, "rb") as f:
                predictions = pickle.load(f)

            with open(os.path.join(raw_dir, "predictions.csv"), "w") as f:
                for image_name in predictions:
                    pred = predictions[image_name]["last_step"]
                    centers = np.stack(
                        [
                            (pred[:, 3] + pred[:, 0]) / 2,
                            (pred[:, 4] + pred[:, 1]) / 2,
                            (pred[:, 5] + pred[:, 2]) / 2,
                        ],
                        axis=1,
                    )
                    centers *= args.downscale
                    centers = np.rint(centers).astype(int)
                    for i, center in enumerate(centers):
                        f.write(image_name)
                        f.write(",")
                        f.write(str(i))
                        f.write(",")
                        f.write(str(center[2]))
                        f.write(",")
                        f.write(str(center[1]))
                        f.write(",")
                        f.write(str(center[0]))
                        f.write("\n")

            shutil.copyfile(
                os.path.join(args.rootdir, "coordinates.csv"),
                os.path.join(raw_dir, "annotations.csv"),
            )
            annotations = load_annotations(os.path.join(raw_dir, "annotations.csv"))
            annotations[:, 2:] = args.downscale * annotations[:, 2:].astype(float)
            with open(os.path.join(raw_dir, "annotations.csv"), "w", newline="") as f:
                csv.writer(f).writerows(annotations)

            crop(
                raw_dir,
                "cropped",
                crop_size=args.patch_size,
                margin=0,
                pos_ratio=1,
                positive_only=True,
                extension=args.extension,
                csv_name="predictions",  # picking predictions
            )

            create_annotation_image(
                raw_dir,
                os.path.join(raw_dir, "annotations.csv"),
                args.patch_size,
                (0, 255, 0),
                args.extension,
            )
            create_annotation_image(
                raw_dir,
                os.path.join(raw_dir, "predictions.csv"),
                args.patch_size,
                (255, 0, 0),
                args.extension,
            )

    if "train" in args.stages and args.task == "tilt":
        train_tilt(
            args.rootdir,
            args.batch_size,
            args.num_workers,
            args.output_dir,
            args.extension,
            args.selected_particles,
            args.crop_dir,
            args.num_epochs,
            args.lr,
            args.network,
            args.width_factor,
            args.depth_factor,
            args.swa,
            args.swa_lr,
            args.patch_size,
            args.num_pos_samples,
            args.shuffle,
            args.augment,
            args.epoch_size,
            args.ae,
            args.eta,
            args.learn_eta,
            args.num_features,
            args.hidden_dim,
            args.pretrain,
        )
    if "predict" in args.stages and args.task == "tilt":
        predict_tilt(
            args.rootdir,
            args.network,
            args.checkpoint,
            args.crop_dir,
            args.dim,
            args.batch_size,
            args.extension,
            args.testset_size,
            args.image_name,
            args.output_dir,
        )
    if "evaluate" in args.stages and args.task == "tilt":
        evaluate_tilt(
            args.predictions,
            args.annotations,
            args.image_name,
            args.output_dir,
            args.selected_particles,
        )
    if args.mail:
        mail_at_the_end(args)


if __name__ == "__main__":
    main(parse_args())



from __future__ import annotations

from typing import TYPE_CHECKING

from spfluo.refinement.refinement import reconstruction_L2
from spfluo.utils.array import array_namespace, get_device
from spfluo.utils.transform import symmetrize_poses
from spfluo.utils.volume import center_of_mass

if TYPE_CHECKING:
    from spfluo.utils.array import Array


def symmetrize(
    particle: "Array",
    center: tuple[float, float],
    symmetry: int,
    psf: "Array",
    lambda_: "Array",
):
    xp = array_namespace(particle, psf, lambda_)
    zc, _, _ = center_of_mass(particle)
    pose_syms = symmetrize_poses(
        xp.asarray(
            [0, 0, 0, -(particle.shape[0] / 2 - zc), center[0], center[1]],
            device=get_device(particle),
            dtype=xp.float64,
        ),
        symmetry=symmetry,
    )
    return reconstruction_L2(
        xp.astype(particle[None], xp.float32),
        xp.astype(psf, xp.float32),
        pose_syms[:, None],
        lambda_,
        symmetry=True,
        device=get_device(particle),
    )[0]
